from typing import Iterable, TypeVar

T = TypeVar('T')


def is_notebook() -> bool:
    """
    Check if it is running in a terminal or in a notebook (or interactive Python).

    Returns
    -------
    bool
        True if it is running in an IPython, False otherwise.
    """
    try:
        from IPython import get_ipython
        return get_ipython() is not None
    except ImportError:
        return False


def get_optional_iterator(**kwargs) -> T:
    """
    Get optional bar progress with or without iterator.

    Parameters
    ----------
    See tqdm documentation.

    Returns
    -------
    tqdm
        Instance of tqdm in case library is installed.

    See Also
    --------
    tqdm : https://tqdm.github.io/docs/tqdm/#tqdm-objects
    """
    iterator = kwargs.pop("iterator", None)
    if is_notebook():
        try:
            from tqdm.notebook import tqdm
            if iterator is None:
                return tqdm(**kwargs)
            else:
                return tqdm(iterator, **kwargs)
        except ImportError:
            return iterator
    else:
        try:
            from tqdm import tqdm
            if iterator is None:
                return tqdm(**kwargs)
            else:
                return tqdm(iterator, **kwargs)
        except ImportError:
            return iterator


def iterate_with_bar(iterator: Iterable[T], **kwargs) -> T:
    """
    Get a tqdm instance iterator or just the iterator.

    Parameters
    ----------
    iterator : Iterable
        Python object to be iterated.

    Other Parameters
    ----------------
    See tqdm documentation.

    Returns
    -------
    tqdm
        Instance of tqdm in case library is installed.

    See Also
    --------
    tqdm : https://tqdm.github.io/docs/tqdm/#tqdm-objects
    """
    return get_optional_iterator(iterator = iterator, **kwargs)


class OptionalTqdm:
    """
    Optional TQDM progress bar. Support all method in case tqdm is not installed.
    """
    def __init__(self, **kwargs) -> None:
        self.__tqdm__ = get_optional_iterator(**kwargs)
        return

    def set_descriptor(self, **kwargs) -> None:
        """
        Set descriptor for tqdm progress bar.

        Other Parameters
        ----------------
        See tqdm documentation.

        See Also
        --------
        tqdm : https://tqdm.github.io/docs/tqdm/#set_description
        """
        if self.__tqdm__ is not None:
            self.__tqdm__.set_description(**kwargs)
        return

    def update(self, **kwargs) -> None:
        """
        Update tqdm progress bar.

        Other Parameters
        ----------------
        See tqdm documentation.

        See Also
        --------
        tqdm : https://tqdm.github.io/docs/tqdm/#update
        """
        if self.__tqdm__ is not None:
            self.__tqdm__.update(**kwargs)
        return

    def close(self) -> None:
        """
        Close tqdm progress bar.

        Other Parameters
        ----------------
        See tqdm documentation.

        See Also
        --------
        tqdm : https://tqdm.github.io/docs/tqdm/#close
        """
        if self.__tqdm__ is not None:
            self.__tqdm__.close()
        return

    def set_postfix(self, **kwargs) -> None:
        """
        Set postfix for tqdm progress bar.

        Other Parameters
        ----------------
        See tqdm documentation.

        See Also
        --------
        tqdm : https://tqdm.github.io/docs/tqdm/#set_postfix
        """
        if self.__tqdm__ is not None:
            self.__tqdm__.set_postfix(**kwargs)
        return

    def reset(self, **kwargs) -> None:
        """
        Reset tqdm progress bar.

        Other Parameters
        ----------------
        See tqdm documentation.

        See Also
        --------
        tqdm : https://tqdm.github.io/docs/tqdm/#reset
        """
        if self.__tqdm__ is not None:
            self.__tqdm__.reset(**kwargs)
        return



from __future__ import annotations

import re
from typing import Dict, Tuple, List

from lxml import etree as et

from xml_common.utils import LengthUnit, GPolygon, GRing
from eml.types import EMLObject, Scope


class GeographicCoverage(EMLObject):
    """
    Geographic coverage information.

    Parameters
    ----------
    description : str
        Short text description of the geographic areal domain of the data set.
    west_bounding : float
        Western-most limit of a bounding box, expressed in degrees of longitude.
    east_bounding : float
        Eastern-most limit of a bounding box, expressed in degrees of longitude.
    north_bounding : float
        Northern-most limit of a bounding box, expressed in degrees of latitude.
    south_bounding : float
        Southern-most limit of a bounding box, expressed in degrees of latitude.
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    altitude_bounding : Tuple[float, float], optional
        The vertical limits of a data set expressed by altitude.
    altitude_units : LengthUnit, optional
        The unit of altitude.
    g_polygon : List[Tuple[List[Tuple[float, float]], List[Tuple[float, float]]]], optional
        This construct creates a spatial ring with a hollow center.
    """
    PRINCIPAL_TAG = "geographicCoverage"

    def __init__(
            self,
            description: str,
            west_bounding: float,
            east_bounding: float,
            north_bounding: float,
            south_bounding: float,
            _id: str = None,
            scope: Scope = None,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            altitude_bounding: Tuple[float, float] = None,
            altitude_units: LengthUnit = None,
            g_polygon: List[GPolygon] = None,
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        self.__description__ = description
        self.__w__ = west_bounding
        self.__e__ = east_bounding
        self.__n__ = north_bounding
        self.__s__ = south_bounding
        if altitude_bounding is not None:
            self.__max_alt__ = altitude_bounding[0]
            self.__min_alt__ = altitude_bounding[1]
        else:
            self.__max_alt__, self.__min_alt__ = None, None
        if altitude_units is not None:
            self.__alt_units__ = altitude_units
        else:
            self.__alt_units__ = None
        self.__polygons__ = list()
        if g_polygon is not None:
            self.__polygons__.extend(g_polygon)
        return

    @property
    def description(self) -> str:
        """str: Short text description of the geographic areal domain of the data set."""
        return self.__description__

    @property
    def west_bounding(self) -> float:
        """float: Western-most limit of a bounding box, expressed in degrees of longitude."""
        return self.__w__

    @property
    def east_bounding(self) -> float:
        """float: Eastern-most limit of a bounding box, expressed in degrees of longitude."""
        return self.__e__

    @property
    def north_bounding(self) -> float:
        """float: Northern-most limit of a bounding box, expressed in degrees of latitude."""
        return self.__n__

    @property
    def south_bounding(self) -> float:
        """float: Southern-most limit of a bounding box, expressed in degrees of latitude."""
        return self.__s__

    @property
    def max_altitude_bounding(self) -> float:
        """float: The maximum altitude extent of coverage."""
        return self.__max_alt__

    @property
    def min_altitude_bounding(self) -> float:
        """float: The minimum altitude extent of coverage."""
        return self.__min_alt__

    @property
    def altitude_units_bounding(self) -> LengthUnit:
        """LengthUnit: The unit of altitude."""
        return self.__alt_units__

    @property
    def g_polygon(self) -> List[GPolygon]:
        """List[GPolygon]: This construct creates a spatial ring with a hollow center."""
        return self.__polygons__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> GeographicCoverage:
        """
        Generate an GeographicCoverage referencing another GeographicCoverage.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        GeographicCoverage
            Object parsed that reference another.
        """
        references = element.find("references", nmap)
        return GeographicCoverage(
            description="",
            west_bounding=float("nan"),
            east_bounding=float("nan"),
            north_bounding=float("nan"),
            south_bounding=float("nan"),
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None),
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> GeographicCoverage:
        """
        Generate a GeographicCoverage that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        GeographicCoverage
            Object parsed.
        """
        description_elem = element.find("geographicDescription", nmap)
        bounding_elem = element.find("boundingCoordinates", nmap)
        if bounding_elem is None:
            raise ValueError("Bounding Coordinates is required.")
        west_bounding_elem = bounding_elem.find("westBoundingCoordinate", nmap)
        west = float(west_bounding_elem.text)
        east_bounding_elem = bounding_elem.find("eastBoundingCoordinate", nmap)
        east = float(east_bounding_elem.text)
        north_bounding_elem = bounding_elem.find("northBoundingCoordinate", nmap)
        north = float(north_bounding_elem.text)
        south_bounding_elem = bounding_elem.find("southBoundingCoordinate", nmap)
        south = float(south_bounding_elem.text)
        alt_bounding_elem = bounding_elem.find("boundingAltitudes", nmap)
        if alt_bounding_elem is not None:
            min_alt_elem = alt_bounding_elem.find("altitudeMinimum", nmap)
            max_alt_elem = alt_bounding_elem.find("altitudeMaximum", nmap)
            altitude_bounding = float(max_alt_elem.text), float(min_alt_elem.text)
            unit_elem = alt_bounding_elem.find("altitudeUnits", nmap)
            unit = LengthUnit.METER
            if unit_elem is not None:
                for candid_unit in LengthUnit:
                    if candid_unit.name.lower() == unit_elem.text.lower():
                        unit = candid_unit
                        break
        else:
            altitude_bounding = None
            unit = None
        polygons = list()
        for polygon_elem in element.findall("datasetGPolygon", nmap):
            polygons.append(cls.xml_to_polygon(polygon_elem, nmap))
        return GeographicCoverage(
            description=description_elem.text if description_elem.text is not None else None,
            west_bounding=west,
            east_bounding=east,
            north_bounding=north,
            south_bounding=south,
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            altitude_bounding=altitude_bounding,
            altitude_units=unit,
            g_polygon=polygons
        )

    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the object.

        Returns
        -------
        lxml.tree.Element
            Geographic Coverage in the Element format.
        """
        geo_element = super().to_element()
        geo_element = self._to_element_(geo_element)
        if not self.referencing:
            desc_elem = self.object_to_element("geographicDescription")
            desc_elem.text = self.description
            geo_element.append(desc_elem)
            if (
                self.west_bounding is not None and
                self.east_bounding is not None and
                self.north_bounding is not None and
                self.south_bounding
            ):
                bounding_elem = self.object_to_element("boundingCoordinates")
                west_elem = self.object_to_element("westBoundingCoordinate")
                west_elem.text = str(self.west_bounding)
                bounding_elem.append(west_elem)
                east_elem = self.object_to_element("eastBoundingCoordinate")
                east_elem.text = str(self.east_bounding)
                bounding_elem.append(east_elem)
                north_elem = self.object_to_element("northBoundingCoordinate")
                north_elem.text = str(self.north_bounding)
                bounding_elem.append(north_elem)
                south_elem = self.object_to_element("southBoundingCoordinate")
                south_elem.text = str(self.south_bounding)
                bounding_elem.append(south_elem)
                if (
                    self.max_altitude_bounding is not None and
                    self.min_altitude_bounding is not None and
                    self.altitude_units_bounding is not None
                ):
                    alt_elem = self.object_to_element("boundingAltitudes")
                    min_alt = self.object_to_element("altitudeMinimum")
                    min_alt.text = str(self.min_altitude_bounding)
                    alt_elem.append(min_alt)
                    max_alt = self.object_to_element("altitudeMaximum")
                    max_alt.text = str(self.max_altitude_bounding)
                    alt_elem.append(max_alt)
                    unit_alt = self.object_to_element("altitudeUnits")
                    unit_alt.text = self.altitude_units_bounding.valid_name
                    alt_elem.append(unit_alt)
                    bounding_elem.append(alt_elem)
                geo_element.append(bounding_elem)
            for polygon in self.g_polygon:
                polygon_elem = self.polygon_to_xml(polygon)
                geo_element.append(polygon_elem)
        return geo_element

    @staticmethod
    def xml_to_polygon(element: et.Element, nmap: Dict) -> GPolygon:
        """
        Convert a lxml.tree.Element into a GPolygon object.

        Parameters
        ----------
        element : lxml.tree.Element
            XML element to be parsed into geographical coordinates
        nmap : Dict
            Namespace.

        Returns
        -------
        GPolygon
            A valid polygon.
        """
        outer_elem = element.find("datasetGPolygonOuterGRing")
        outer = GeographicCoverage.xml_to_ring(outer_elem, nmap)
        exclusion = list()
        for excl_elem in element.findall("datasetGPolygonExclusionGRing", nmap):
            exclusion.append(GeographicCoverage.xml_to_ring(excl_elem, nmap))
        return GPolygon(outer, exclusion=exclusion)

    @staticmethod
    def polygon_to_xml(polygon: GPolygon) -> et.Element:
        """
        Convert a GPolygon object into an XML Element.

        Parameters
        ----------
        polygon : GPolygon
            A valid GPolygon object.

        Returns
        -------
        lxml.tree.Element
            XML element with the GPolygon information.
        """
        element = et.Element("datasetGPolygon")
        outer_elem = et.Element("datasetGPolygonOuterGRing")
        outer_elem.append(GeographicCoverage.ring_to_xml(polygon.outer))
        element.append(outer_elem)
        for excl in polygon.exclusion:
            excl_elem = et.Element("datasetGPolygonExclusionGRing")
            excl_elem.append(GeographicCoverage.ring_to_xml(excl))
            element.append(excl_elem)
        return element

    @staticmethod
    def xml_to_ring(element: et.Element, nmap: Dict) -> GRing:
        """
        Convert a lxml.tree.Element into a GRing object.

        Parameters
        ----------
        element : lxml.tree.Element
            XML element to be parsed into geographical coordinates.
        nmap : Dict
            Namespace.

        Returns
        -------
        GRing
            A valid ring.
        """
        g_ring = element.find("gRing", nmap)
        g_ring_points = element.findall("gRingPoint", nmap)
        points = list()
        if g_ring is not None:
            for point in re.split(r'(?<!,)\s', g_ring.text):
                lat, long = point.split(",")
                points.append((float(lat.strip()), float(long.strip())))
        elif len(g_ring_points) > 0:
            for point in g_ring_points:
                lat = float(point.find("gRingLatitude", nmap).text)
                long = float(point.find("gRingLongitude", nmap).text)
                points.append((lat, long))
        return GRing(*points)

    @staticmethod
    def ring_to_xml(ring: GRing) -> et.Element:
        """
        Convert a GRing object into an XML Element.

        Parameters
        ----------
        ring : GRing
            A valid GRing object.

        Returns
        -------
        lxml.tree.Element
            XML element with the GPolygon information.
        """
        element = et.Element("gRing")
        element.text = " ".join(
            [f"{lat}, {long}" for lat, long in ring]
        )
        return element



from abc import ABC
from typing import Dict

from lxml import etree as et

from xml_common import XMLObject


class _NoTagObject(XMLObject, ABC):
    """
    Abstract class of object with no tag on EML schema
    """
    def __init__(self) -> None:
        super().__init__()
        self.__tag__ = None
        return

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the Responsible Party object.

        Returns
        -------
        lxml.tree.Element
            Object in the XML Element format.

        Raises
        ------
        RuntimeError
            If tag is not set before calling this method.
        """
        if self.__tag__ is None:
            raise RuntimeError("First, set tag to call `to_element`")
        return et.Element(self.__tag__)

    @classmethod
    def check_principal_tag(cls, tag: str, nmap: Dict) -> None:
        """
        Always return True due tag not existing until set

        Parameters
        ----------
        tag : str
            Any tag.
        nmap : Dict
            Namespace.
        """
        return

    def set_tag(self, tag: str) -> None:
        """
        Set tag for the element output.

        Parameters
        ----------
        tag : str
            Tag of the XML output element.
        """
        self.__tag__ = tag
        return



from dwca.classes.data_file import DataFile, DataFileType
from dwca.classes.occurrence import Occurrence
from dwca.classes.organism import Organism
from dwca.classes.material_entity import MaterialEntity
from dwca.classes.material_sample import MaterialSample
from dwca.classes.event import Event
from dwca.classes.location import Location
from dwca.classes.geological_context import GeologicalContext
from dwca.classes.identification import Identification
from dwca.classes.taxon import Taxon
from dwca.classes.resource_relationship import ResourceRelationship
from dwca.classes.measurement_or_fact import MeasurementOrFact
from dwca.classes.chronometric_age import ChronometricAge
from dwca.classes.outside_class import OutsideClass



import os
import tempfile
import unittest
import zipfile

from dwca.base import DarwinCoreArchive
from dwca.classes import OutsideClass
from dwca.terms import OutsideTerm
from eml import EML
from eml.resources import EMLResource
from eml.types import ResponsibleParty, IndividualName
from test_xml.test_xml import TestXML
from xml_common.utils import Language

PATH = os.path.abspath(os.path.dirname(__file__))


class TestDWCACommon(TestXML):
    def setUp(self) -> None:
        self.object = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "example_archive.zip")
        )
        return

    def __test_minimal__(self):
        minimal_dwca = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "minimal.zip")
        )
        self.assertIsNone(minimal_dwca.metadata, "Metadata file generated from nowhere.")
        self.assertIsNone(minimal_dwca.metadata_filename, "Metadata file (name) generated from nowhere.")
        self.assertEqual(0, len(minimal_dwca.extensions), "Extension from nowhere.")
        self.assertEqual("minimal_core.txt", minimal_dwca.core.filename, "Wrong filename in core.")
        minimal_dwca.core = OutsideClass(0, "http://www.example.org/minimalCore", "minimal_core.txt", [
            OutsideTerm(0, "http://www.example.org/minimalTerm")
        ])

    def __test_attributes__(self):
        self.assertIsNotNone(self.object.metadata, "Doesn't have metadata")
        self.assertEqual("eml.xml", self.object.metadata_filename, "Wrong file name of metadata")
        self.assertEqual(3, len(self.object.extensions), "Missing or wrong number of extensions")
        self.assertEqual("taxon.txt", self.object.core.filename, "Wrong filename in Core")
        self.assertEqual("Core:"
                         "\n\tclass: http://rs.tdwg.org/dwc/terms/Taxon"
                         "\n\tfilename: taxon.txt"
                         "\n\tcontent: 163460 entries", str(self.object.core), "Wrong string of core")
        self.assertCountEqual(
            ["speciesprofile.txt", "reference.txt", "identification.txt"],
            [extension.filename for extension in self.object.extensions],
            "Extension not read"
        )
        self.assertEqual("Extension:"
                         "\n\tclass: http://rs.gbif.org/terms/1.0/SpeciesProfile"
                         "\n\tfilename: speciesprofile.txt"
                         "\n\tcontent: 153621 entries", str(self.object.extensions[0]),
                         "Wrong string of extension (species profile)")
        self.assertEqual("Extension:"
                         "\n\tclass: http://rs.gbif.org/terms/1.0/Reference"
                         "\n\tfilename: reference.txt"
                         "\n\tcontent: 98518 entries", str(self.object.extensions[1]),
                         "Wrong string of extension (reference)")
        self.assertEqual("Extension:"
                         "\n\tclass: http://rs.tdwg.org/dwc/terms/Identification"
                         "\n\tfilename: identification.txt"
                         "\n\tcontent: 7617 entries", str(self.object.extensions[2]),
                         "Wrong string of extension (identification)")
        for ext in self.object.extensions:
            self.assertEqual(
                "http://rs.tdwg.org/dwc/terms/taxonID",
                ext.fields[0],
                f"Core id not set in extension {ext.uri}"
            )
        self.assertEqual(Language.ENG, self.object.language, "Wrong language")

    def __test_merge__(self):
        dataset1 = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "dataset_1.zip")
        )
        dataset2 = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "dataset_2.zip")
        )
        original_expected = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "merged_dataset.zip")
        )
        metadata = EML(
            package_id="merge_data",
            system="http://gbif.org",
            resource_type=EMLResource.DATASET
        )
        metadata.initialize_resource(
            "Merged Dataset",
            ResponsibleParty(
                individual_name=IndividualName("Doe", "John")
            ),
            contact=[ResponsibleParty(
                individual_name=IndividualName("Doe", "Jane")
            )],
        )
        actual_12 = DarwinCoreArchive.merge(
            dataset1, dataset2, eml=metadata
        )
        actual_21 = DarwinCoreArchive.merge(
            dataset2, dataset1, eml=metadata
        )
        actual_121 = DarwinCoreArchive.merge(
            actual_12, dataset1, eml=metadata
        )
        actual_112 = DarwinCoreArchive.merge(
            dataset1, actual_12, eml=metadata
        )
        new_expected = DarwinCoreArchive.merge(
            original_expected, dataset1, eml=metadata
        )
        for actual, expected, name in [
            (actual_12, original_expected, "1 + 2"),
            (actual_21, original_expected, "2 + 1"),
            (actual_121, new_expected, "1 + 2 + 1"),
            (actual_112, new_expected, "1 + 1 + 2"),
        ]:
            self.assertEqualTree(
                actual.metadata.to_element(),
                expected.metadata.to_element(),
                f"Incorrect EML set ({name})"
            )
            self.assertEqual(
                len(expected.core), len(actual.core), f"Wrong number of values on core ({name})"
            )
            self.assertEqual(len(expected.extensions), len(actual.extensions), f"Not the same extensions ({name})")
            for exp_ext, act_ext in zip(expected.extensions, actual.extensions):
                self.assertEqual(
                    len(exp_ext), len(act_ext), f"Wrong number of values on extension: {act_ext} ({name})"
                )
            self.assertDictEqual(
                expected.dataset_metadata,
                actual.dataset_metadata,
                f"Different metadata of datasets ({name})"
            )
            with tempfile.NamedTemporaryFile("wb") as file:
                actual.to_file(file.name)
                with zipfile.ZipFile(file.name, "r") as zip_file:
                    self.assertEqual(5 + len(actual.dataset_metadata), len(zip_file.namelist()), "Error writing DwC-A file with metadata.")
                    datasets = ["eml.xml"]
                    for dataset in actual.dataset_metadata.keys():
                        if dataset != "metadata":
                            datasets.append("dataset/" + dataset + ".xml")
                    self.assertCountEqual(
                        ['meta.xml', 'taxon.txt', 'reference.txt', 'extension1.txt', 'extension2.txt'] + datasets,
                        zip_file.namelist(),
                        "Incorrect file saved"
                    )


if __name__ == '__main__':
    unittest.main()



import logging
import os
import unittest
from lxml import etree as et

from eml.types import ResponsibleParty, EMLAddress, IndividualName, PositionName, OrganizationName
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestResponsibleParty(TestXML):
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        "scope": "document",
        "phonetype": "voice",
    }

    def setUp(self) -> None:
        logging.getLogger().setLevel(logging.DEBUG)
        with open(os.path.join(PATH, os.pardir, "example_data", "eml.xml"), "r", encoding="utf-8") as file:
            content = file.read()
        base_file = et.fromstring(content)
        self.nmap = base_file.nsmap
        self.dataset_xml = base_file.find("dataset", namespaces=base_file.nsmap)
        self.individual = IndividualName(first_name="John", last_name="Doe")
        self.position = PositionName("Digitizer")
        self.organization = OrganizationName("Example Organization")
        return

    def test_parse_creator(self):
        xml = self.dataset_xml.find("creator", namespaces=self.nmap)
        text = et.tostring(xml, pretty_print=True).decode("utf-8")
        resp_party = ResponsibleParty.from_string(text)
        self.assertEqual("Creator Organization", resp_party.organization_name, "Error on parsing organization name")
        self.assertEqual("info@creator.org", resp_party.mail[0], "Error on parsing electronic mail address")
        self.assertEqual("https://example.creator.org", resp_party.url[0], "Error on parsing online url")
        self.assertRaises(
            RuntimeError,
            resp_party.to_element
        )
        resp_party.set_tag("creator")
        self.assertEqualTree(et.fromstring(text), resp_party.to_element(), "Creator error on to element")

    def test_parse_metadata_provider(self):
        xml = self.dataset_xml.find("metadataProvider", namespaces=self.nmap)
        text = et.tostring(xml, pretty_print=True).decode("utf-8")
        resp_party = ResponsibleParty.from_string(text)
        self.assertEqual(
            "Metadata Provider Organization",
            resp_party.organization_name,
            "Error on parsing organization name"
        )
        self.assertEqual(
            "Metadata Manager",
            resp_party.position_name,
            "Error on parsing position name"
        )
        self.assertEqual("info@metadata.org", resp_party.mail[0], "Error on parsing electronic mail address")
        self.assertEqual(
            EMLAddress(city="Ostend", postal_code="4568", country="BE"),
            resp_party.address[0],
            "Error on parsing address"
        )
        self.assertEqual("https://example.metadata.org", resp_party.url[0], "Error on parsing online url")
        resp_party.set_tag("metadataProvider")
        self.assertEqualTree(et.fromstring(text), resp_party.to_element(), "Metadata Provider error on to element")

    def test_parse_associated_party(self):
        xml = self.dataset_xml.find("associatedParty", namespaces=self.nmap)
        text = et.tostring(xml, pretty_print=True).decode("utf-8")
        resp_party = ResponsibleParty.from_string(text)
        self.assertEqual("Doe, J.", str(resp_party.individual_name), "Error on parsing individual name")
        self.assertEqual(
            "Custodian Steward",
            resp_party.organization_name,
            "Error on parsing organization name"
        )
        self.assertEqual("info@custodian.org", resp_party.mail[0], "Error on parsing electronic mail address")
        self.assertEqual(EMLAddress(country="Belgium"), resp_party.address[0], "Error on parsing address")
        self.assertEqual("486554390", resp_party.phone[0], "Error on parsing phone")
        resp_party.set_tag("associatedParty")
        text = text.replace("<role>custodianSteward</role>", "")
        self.assertEqualTree(et.fromstring(text), resp_party.to_element(), "Wrong to element")

    def test_reference_resp_party(self):
        reference_xml = """
<creator>
    <references>1</references>
</creator>
        """
        eml_resp_party = ResponsibleParty.from_string(reference_xml)
        self.assertEqual("1", eml_resp_party.id, "Id set incorrectly")
        self.assertTrue(eml_resp_party.referencing, "Referencing wrong initialized")
        self.assertIsNone(
            eml_resp_party.references.system,
            "References system wrong initialized"
        )
        eml_resp_party.set_tag("creator")
        self.assertEqualTree(
            et.fromstring(reference_xml),
            eml_resp_party.to_element(),
            "Wrong referrer to element"
        )

    def test_reference_system(self):
        reference_xml = """
<creator>
    <references system="http://gbif.org">1</references>
</creator>
        """
        eml_resp_party = ResponsibleParty.from_string(reference_xml)
        self.assertEqual("1", eml_resp_party.id, "Id set incorrectly")
        self.assertTrue(eml_resp_party.referencing, "Referencing wrong initialized")
        self.assertEqual(
            "http://gbif.org",
            eml_resp_party.references.system,
            "References system wrong initialized"
        )
        self.assertRaises(
            RuntimeError,
            eml_resp_party.to_element
        )
        eml_resp_party.set_tag("creator")
        self.assertEqualTree(
            et.fromstring(reference_xml),
            eml_resp_party.to_element(),
            "Wrong referrer to element"
        )

    def test_invalid_reference(self):
        reference_xml = """
<creator id="10">
    <references>1</references>
</creator>
        """
        self.assertRaises(
            AssertionError,
            ResponsibleParty.from_string,
            reference_xml
        )

    def test_str_individual(self):
        just_individual = ResponsibleParty(individual_name=self.individual)
        self.assertEqual("Doe, J.", str(just_individual), "Error on string of individual.")
        individual_position = ResponsibleParty(
            individual_name=self.individual,
            position_name=self.position
        )
        self.assertEqual(
            "Doe, J. (Digitizer)",
            str(individual_position),
            "Error on string of individual with position."
        )
        individual_organization = ResponsibleParty(
            individual_name=self.individual,
            organization_name=self.organization
        )
        self.assertEqual(
            "Doe, J. (Example Organization)",
            str(individual_organization),
            "Error on string of individual with organization."
        )
        individual_complete = ResponsibleParty(
            individual_name=self.individual,
            position_name=self.position,
            organization_name=self.organization
        )
        self.assertEqual(
            "Doe, J. (Digitizer at Example Organization)",
            str(individual_complete),
            "Error on string of complete individual."
        )

    def test_str_position(self):
        just_position = ResponsibleParty(position_name=self.position)
        self.assertEqual("Digitizer", str(just_position), "Error on string of position.")
        position_organization = ResponsibleParty(
            position_name=self.position,
            organization_name=self.organization
        )
        self.assertEqual(
            "Digitizer at Example Organization",
            str(position_organization),
            "Error on string of position with organization."
        )

    def test_str_organization(self):
        just_organization = ResponsibleParty(organization_name=self.organization)
        self.assertEqual("Example Organization", str(just_organization), "Error on string of organization.")


if __name__ == '__main__':
    unittest.main()



from xml_common.utils import CamelCaseEnum


class EstablishmentMeans(CamelCaseEnum):
    """
    Terms intended to be used as a controlled value for Darwin Core terms with local name establishmentMeans.

    For details and rationale, see Groom et al. 2019. Improving Darwin Core for research
    and management of alien species. https://doi.org/10.3897/biss.3.38084

    Attributes
    ----------
    NATIVE : int
        A taxon occurring within its natural range.
    NATIVE_REINTRODUCED : int
        A taxon re-established by direct introduction into its natural range, but from where it had become extinct.
    INTRODUCED : int
        Establishment of a taxon by human agency into an area that is not part of its natural range.
    INTRODUCED_ASSISTED_COLONISATION : int
        Establishment of a taxon specifically with the intention of creating a self-sustaining wild population.
    VAGRANT : int
        The temporary occurrence of a taxon far outside its natural or migratory range.
    UNCERTAIN : int
        The origin of the occurrence of the taxon in an area is obscure.
    """
    NATIVE = 1
    NATIVE_REINTRODUCED = 2
    INTRODUCED = 3
    INTRODUCED_ASSISTED_COLONISATION = 4
    VAGRANT = 5
    UNCERTAIN = 6

    def get_iri(self) -> str:
        """
        Generate the Term IRI (Internationalized Resource Identifier) of the Establishment Means.

        Returns
        -------
        str
            Term IRI.
        """
        return f"http://rs.tdwg.org/dwcem/values/e{self.value:03}"



import unittest

from lxml import etree as et

from xml_common.utils import Language
from test_xml.test_xml import TestXML
from eml.types import IndividualName, I18nString


class TestPerson(TestXML):
    def test_individual_name_last_name(self):
        variable = IndividualName("Saez")
        self.assertEqual("Saez", str(variable), "Not equal to string")
        self.assertEqual(
            "<Individual Name {Last Name: Saez [eng]}>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")

    def test_individual_name_first_name(self):
        variable = IndividualName("Saez", first_name="Juan")
        self.assertEqual("Saez, J.", str(variable), "Not equal to string")
        self.assertEqual(
            "<Individual Name {Last Name: Saez [eng], First Name: Juan [eng]}>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")
        self.assertEqual(Language.ENG, variable.first_name[0].language, "Language set incorrectly")

    def test_individual_name_salutation(self):
        variable = IndividualName("Saez", first_name="Juan", salutation="MSc")
        self.assertEqual("Saez, J. MSc.", str(variable), "Not equal to string")
        self.assertEqual(
            "<Individual Name {Last Name: Saez [eng], First Name: Juan [eng], Salutation: MSc [eng]}>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")
        self.assertEqual(Language.ENG, variable.first_name[0].language, "Language set incorrectly")
        self.assertEqual(Language.ENG, variable.salutation[0].language, "Language set incorrectly")

    def test_individual_name_spanish(self):
        variable = IndividualName("Saez", first_name="Juan", salutation="MSc", language=Language.ESP)
        self.assertEqual("Saez, J. MSc.", str(variable), "Not equal to string")
        self.assertEqual(
            "<Individual Name {Last Name: Saez [esp], First Name: Juan [esp], Salutation: MSc [esp]}>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ESP, variable.last_name.language, "Language set incorrectly")
        self.assertEqual(Language.ESP, variable.first_name[0].language, "Language set incorrectly")
        self.assertEqual(Language.ESP, variable.salutation[0].language, "Language set incorrectly")

    def test_individual_name_multi_language(self):
        variable = IndividualName(
            "Saez", first_name=[
                I18nString("Juan", "eng"),
                I18nString("Juan", "esp")
            ], salutation=[
                I18nString("MSc", "eng"),
                I18nString("Don", "esp")
            ]
        )
        self.assertRaises(ValueError, str, variable)
        self.assertEqual("Saez, J. MSc.", variable.lang(Language.ENG), "Not equal to string")
        self.assertEqual("Saez, J. Don.", variable.lang(Language.ESP), "Not equal to string")
        self.assertEqual(
            "<Individual Name {Last Name: Saez [eng], "
            "First Name: Juan [eng]/Juan [esp], "
            "Salutation: MSc [eng]/Don [esp]}>",
            repr(variable),
            "Not equal to string"
        )
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")
        self.assertCountEqual(
            [Language.ENG, Language.ESP],
            [first_name.language for first_name in variable.first_name],
            "Language set incorrectly"
        )
        self.assertCountEqual(
            [Language.ENG, Language.ESP],
            [salutation.language for salutation in variable.salutation],
            "Language set incorrectly"
        )

    def test_parse(self):
        self.assertIsNone(IndividualName.parse(None, {}), "Parsing something")
        example_xml = """
<individualName>
    <surName>Saez</surName>
</individualName>
        """
        variable = IndividualName.from_string(example_xml)
        self.assertEqual("Saez", str(variable), "Not equal to string")
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")

    def test_parse_complete(self):
        self.assertIsNone(IndividualName.parse(None, {}), "Parsing something")
        example_xml = """
<individualName>
    <surName>Saez</surName>
    <givenName xml:lang="esp">Juan</givenName>
    <salutation xml:lang="esp">MSc</salutation>
</individualName>
        """
        variable = IndividualName.from_string(example_xml)
        self.assertRaises(ValueError, str, variable)
        self.assertEqual("Saez", variable.lang(Language.ENG), "Not equal to string")
        self.assertEqual("Saez, J. MSc.", variable.lang(Language.ESP), "Not equal to string")
        self.assertEqual(Language.ENG, variable.last_name.language, "Language set incorrectly")
        self.assertEqual(Language.ESP, variable.first_name[0].language, "Language set incorrectly")
        self.assertEqual(Language.ESP, variable.salutation[0].language, "Language set incorrectly")

    def test_to_element(self):
        variable = IndividualName("Saez", language=Language.ESP)
        expected = et.Element("individualName")
        sur_name = et.Element("surName")
        sur_name.text = "Saez"
        sur_name.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(sur_name)
        self.assertEqualTree(expected, variable.to_element(), "Error on element")
    
    def test_to_element_complete(self):
        variable = IndividualName("Saez", first_name="Juan", salutation="Don", language=Language.ESP)
        expected = et.Element("individualName")
        sur_name = et.Element("surName")
        sur_name.text = "Saez"
        sur_name.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(sur_name)
        first_name = et.Element("givenName")
        first_name.text = "Juan"
        first_name.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(first_name)
        salutation = et.Element("salutation")
        salutation.text = "Don"
        salutation.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(salutation)
        self.assertEqualTree(expected, variable.to_element(), "Error on complete element")

    def test_to_element_multilanguage(self):
        variable = IndividualName(
            "Saez",
            first_name=[
                I18nString("Juan", Language.ESP),
                I18nString("John", Language.ENG),
            ],
            salutation=[
                I18nString("Don", Language.ESP),
                I18nString("Mr.", Language.ENG),
            ],
            language=Language.ESP
        )
        expected = et.Element("individualName")
        sur_name = et.Element("surName")
        sur_name.text = "Saez"
        sur_name.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(sur_name)
        first_name_1 = et.Element("givenName")
        first_name_1.text = "Juan"
        first_name_1.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(first_name_1)
        first_name_2 = et.Element("givenName")
        first_name_2.text = "John"
        first_name_2.set("{http://www.w3.org/XML/1998/namespace}lang", "eng")
        expected.append(first_name_2)
        salutation_1 = et.Element("salutation")
        salutation_1.text = "Don"
        salutation_1.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        expected.append(salutation_1)
        salutation_2 = et.Element("salutation")
        salutation_2.text = "Mr."
        salutation_2.set("{http://www.w3.org/XML/1998/namespace}lang", "eng")
        expected.append(salutation_2)
        self.assertEqualTree(expected, variable.to_element(), "Error on multilanguage element")

    def test_lang(self):
        variable = IndividualName("Saez")
        self.assertRaises(AssertionError, variable.lang, Language.ESP)
        self.assertEqual("Saez", variable.lang(Language.ENG), "Error on simple call to lang")


if __name__ == '__main__':
    unittest.main()



from unittest import main
from lxml import etree as et
from eml.types import ResearchProject, Role, ResponsibleParty, OrganizationName
from test_xml.test_xml import TestXML


class TestResearchProject(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
        "{http://www.w3.org/XML/1998/namespace}lang": "eng"
    }

    def test_parse_none(self):
        self.assertIsNone(ResearchProject.parse(None, {}), "Research Project from nowhere.")

    def test_parse_referenced(self):
        text_xml = """
<project>
    <references>some-id</references>
</project>
        """
        res_proj = ResearchProject.from_string(text_xml)
        self.assertEqual(
            "some-id",
            res_proj.id,
            "Research Project parsed incorrectly."
        )
        self.assertIsNone(res_proj.title, "Title from nowhere.")
        self.assertEqual(0, len(res_proj.personnel), "Personnel from nowhere.")
        self.assertIsNone(res_proj.abstract, "Abstract from nowhere.")
        self.assertIsNone(res_proj.funding, "Funding from nowhere.")
        self.assertEqual(0, len(res_proj.award), "Award from nowhere.")
        self.assertIsNone(res_proj.study_area_description, "Study area description from nowhere.")
        self.assertIsNone(res_proj.design_description, "Design description from nowhere.")
        self.assertEqual(0, len(res_proj.related_project), "Related project from nowhere.")
        res_proj.set_tag("project")
        self.assertEqualTree(et.fromstring(text_xml), res_proj.to_element(), "Error on to element.")

    def test_parse_simple(self):
        text_xml = """
<project>
    <title>Example Title</title>
    <personnel>
        <organizationName>Example Organization</organizationName>
        <role>contentProvider</role>
    </personnel>
</project>
        """
        res_proj = ResearchProject.from_string(text_xml)
        self.assertEqual("Example Title", res_proj.title, "Error parsing title.")
        self.assertEqual(1, len(res_proj.titles), "Error parsing titles.")
        self.assertEqual(1, len(res_proj.personnel), "Error parsing personnel.")
        self.assertEqual(Role.CONTENT_PROVIDER, res_proj.personnel[0][1], "Error parsing personnel role.")
        self.assertIsNone(res_proj.abstract, "Abstract from nowhere.")
        self.assertIsNone(res_proj.funding, "Funding from nowhere.")
        self.assertEqual(0, len(res_proj.award), "Award from nowhere.")
        self.assertIsNone(res_proj.study_area_description, "Study area description from nowhere.")
        self.assertIsNone(res_proj.design_description, "Design description from nowhere.")
        self.assertEqual(0, len(res_proj.related_project), "Related project from nowhere.")
        res_proj.set_tag("project")
        self.assertEqualTree(et.fromstring(text_xml), res_proj.to_element(), "Error on to element.")

    def test_no_title(self):
        text_xml = """
<project>
    <personnel>
        <organizationName>Example Organization</organizationName>
        <role>contentProvider</role>
    </personnel>
</project>
        """
        with self.assertRaisesRegex(TypeError, "one title"):
            ResearchProject.from_string(text_xml)

    def test_no_personnel(self):
        text_xml = """
<project>
    <title>Example Title</title>
</project>
        """
        with self.assertRaisesRegex(TypeError, "one personnel"):
            ResearchProject.from_string(text_xml)

    def test_title_as_string(self):
        res_proj = ResearchProject(
            title="Example Title",
            personnel=[(
                ResponsibleParty(organization_name=OrganizationName("Example Organization")),
                Role.CONTENT_PROVIDER
            )],
        )
        self.assertEqual("Example Title", res_proj.title, "Error parsing title.")
        self.assertEqual(1, len(res_proj.titles), "Error parsing titles.")
        self.assertEqual(1, len(res_proj.personnel), "Error parsing personnel.")
        self.assertEqual(Role.CONTENT_PROVIDER, res_proj.personnel[0][1], "Error parsing personnel role.")
        self.assertIsNone(res_proj.abstract, "Abstract from nowhere.")
        self.assertIsNone(res_proj.funding, "Funding from nowhere.")
        self.assertEqual(0, len(res_proj.award), "Award from nowhere.")
        self.assertIsNone(res_proj.study_area_description, "Study area description from nowhere.")
        self.assertIsNone(res_proj.design_description, "Design description from nowhere.")
        self.assertEqual(0, len(res_proj.related_project), "Related project from nowhere.")

    def test_related_project(self):
        text_xml = """
<project>
    <title>Example Title</title>
    <personnel id="ex_org">
        <organizationName>Example Organization</organizationName>
        <role>contentProvider</role>
    </personnel>
    <relatedProject>
        <title>Related Project Title</title>
        <personnel>
            <references>ex_org</references>
            <role>contentProvider</role>
        </personnel>
    </relatedProject>
</project>
        """
        res_proj = ResearchProject.from_string(text_xml)
        self.assertEqual("Example Title", res_proj.title, "Error parsing title.")
        self.assertEqual(1, len(res_proj.titles), "Error parsing titles.")
        self.assertEqual(1, len(res_proj.personnel), "Error parsing personnel.")
        self.assertEqual(Role.CONTENT_PROVIDER, res_proj.personnel[0][1], "Error parsing personnel role.")
        self.assertIsNone(res_proj.abstract, "Abstract from nowhere.")
        self.assertIsNone(res_proj.funding, "Funding from nowhere.")
        self.assertEqual(0, len(res_proj.award), "Award from nowhere.")
        self.assertIsNone(res_proj.study_area_description, "Study area description from nowhere.")
        self.assertIsNone(res_proj.design_description, "Design description from nowhere.")
        self.assertEqual(1, len(res_proj.related_project), "Error parsing related project.")
        self.assertEqual("ex_org", res_proj.related_project[0].personnel[0][0].id, "Error parsing personnel of related project.")
        res_proj.set_tag("project")
        self.assertEqualTree(et.fromstring(text_xml), res_proj.to_element(), "Error on to element.")

if __name__ == '__main__':
    main()



import os
import unittest

from lxml import etree as et

from eml.base import EMLAdditionalMetadata, EMLMetadata
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestMetadata(TestXML):
    def test_none(self):
        self.assertIsNone(EMLAdditionalMetadata.parse(None, {}), "Additional Metadata from nowhere")
        self.assertIsNone(EMLMetadata.parse(None, {}), "Metadata from nowhere")

    def test_parse(self):
        eml_path = os.path.join(PATH, os.pardir, "example_data", "eml.xml")
        with open(eml_path, "r", encoding="utf-8") as file:
            xml_tree = et.fromstring(file.read()).find("additionalMetadata")
        add_metadata = EMLAdditionalMetadata.from_string(et.tostring(xml_tree).decode("utf-8"))
        metadata_tree = xml_tree.find("metadata")
        metadata = EMLMetadata.from_string(et.tostring(metadata_tree).decode("utf-8"))
        self.assertEqualTree(metadata_tree, metadata.to_element(), "Error on metadata")
        self.assertEqualTree(xml_tree, add_metadata.to_element(), "Error on additional metadata")

    def test_complete(self):
        text_xml = """
<additionalMetadata id="A unique ID">
    <metadata>
        <example>Example</example>
        <anotherExample>Different Example</anotherExample>
    </metadata>
    <describes>Section 1</describes>
    <describes>Dataset 3</describes>
</additionalMetadata>
        """
        add_metadata = EMLAdditionalMetadata.from_string(text_xml)
        self.assertEqualTree(et.fromstring(text_xml), add_metadata.to_element(), "Error on to element")


if __name__ == '__main__':
    unittest.main()



import os
import tempfile
import unittest
import zipfile

from dwca import DarwinCoreArchive
from eml import EML
from eml.resources import EMLResource
from eml.types import ResponsibleParty, IndividualName
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestMultipleDataset(TestXML):

    def test_unique_dataset(self):
        dwc = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "example_archive.zip")
        )
        self.assertEqual(1, len(dwc.dataset_metadata), "Incorrect number of metadata for unique dataset.")
        self.assertEqual(
            dwc.metadata.to_xml(),
            dwc.dataset_metadata["metadata"].to_xml(),
            "Difference between the main metadata and metadata of datasets."
        )

    def test_multiple_dataset(self):
        dwc = DarwinCoreArchive.from_file(
            os.path.join(PATH, os.pardir, "example_data", "merged_dataset.zip")
        )
        self.assertEqual(3, len(dwc.dataset_metadata), "Incorrect number of metadata.")
        self.assertEqual(
            dwc.metadata.to_xml(),
            dwc.dataset_metadata["metadata"].to_xml(),
            "Difference between the main metadata and metadata of datasets."
        )
        self.assertEqual(
            "dataset_1", dwc.dataset_metadata["dataset_1"].package_id,
            "Incorrect `dataset_1` package_id"
        )
        self.assertEqual(
            "Example for Darwin Core Archive 1", dwc.dataset_metadata["dataset_1"].resource.title,
            "Incorrect `dataset_1` title"
        )
        self.assertEqual(
            "dataset_2", dwc.dataset_metadata["dataset_2"].package_id,
            "Incorrect `dataset_2` package_id"
        )
        self.assertEqual(
            "Example for Darwin Core Archive 2", dwc.dataset_metadata["dataset_2"].resource.title,
            "Incorrect `dataset_2` title"
        )

    def test_write_multiple_dataset(self):
        dwc = DarwinCoreArchive("test")
        dwc.generate_eml("eml.xml")
        dwc.metadata.initialize_resource(
            "A Multiple Dataset Test",
            ResponsibleParty(individual_name=IndividualName("Doe", "Example")),
            contact=[ResponsibleParty(individual_name=IndividualName(
                "Doe", "Jane"
            ))]
        )
        dwc.dataset_metadata["dataset_1"] = EML(
            "another_dataset", "http://gbif.org", EMLResource.DATASET
        )
        dwc.dataset_metadata["dataset_1"].initialize_resource(
            "First extra dataset",
            ResponsibleParty(individual_name=IndividualName("Doe", "Example")),
            contact=[ResponsibleParty(individual_name=IndividualName(
                "Doe", "Jane"
            ))]
        )
        dwc.dataset_metadata["dataset_2"] = EML(
            "one_more_dataset", "http://gbif.org", EMLResource.DATASET
        )
        dwc.dataset_metadata["dataset_2"].initialize_resource(
            "Second extra dataset",
            ResponsibleParty(individual_name=IndividualName("Doe", "Example")),
            contact=[ResponsibleParty(individual_name=IndividualName(
                "Doe", "Jane"
            ))]
        )
        with tempfile.NamedTemporaryFile("wb") as file:
            dwc.to_file(file.name)
            with zipfile.ZipFile(file.name, "r") as zipped_file:
                presented_files = zipped_file.namelist()
                self.assertEqual(4, len(presented_files), "Wrong number of files on dataset")
                self.assertCountEqual(
                    ["dataset/dataset_1.xml", "dataset/dataset_2.xml"],
                    list(filter(lambda x: "dataset/" in x, presented_files)),
                    f"Expected datasets not present: {presented_files}"
                )
                dataset1 = EML.from_string(zipped_file.read("dataset/dataset_1.xml").decode())
                self.assertEqualTree(dwc.dataset_metadata["dataset_1"].to_element(), dataset1.to_element(), "Wrong parsed dataset 1")
                dataset2 = EML.from_string(zipped_file.read("dataset/dataset_2.xml").decode())
                self.assertEqualTree(dwc.dataset_metadata["dataset_2"].to_element(), dataset2.to_element(), "Wrong parsed dataset 2")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Dict, Union, List

from lxml import etree as et

from xml_common.utils import Language
from xml_common import XMLObject
from eml.types import I18nString


class IndividualName(XMLObject):
    """
    Class representing the full name of the person being described.

    Parameters
    ----------
    last_name : str or I18nString
        The last name of the individual.
    first_name : str, I18nString or List[I18nString], optional
        The given name of the individual.
    salutation : str, I18nString or List[I18nString], optional
        The salutation used to address an individual.
    language : Language, optional
        In case of given, the language of all the parameters given
    """
    PRINCIPAL_TAG = "individualName"

    def __init__(
            self, last_name: Union[str, I18nString],
            first_name: Union[str, I18nString, List[I18nString]] = None,
            salutation: Union[str, I18nString, List[I18nString]] = None,
            language: Language = None,
    ) -> None:
        super().__init__()
        if language is not None:
            self.__last_name__ = I18nString(last_name, lang=language)
        else:
            self.__last_name__ = I18nString(last_name)
        self.__languages__ = {self.__last_name__.language}
        self.__first_name__: List[I18nString] = list()
        if isinstance(first_name, list):
            self.__first_name__.extend(first_name)
        elif first_name is not None:
            if language is not None:
                self.__first_name__.append(I18nString(first_name, lang=language))
            else:
                self.__first_name__.append(I18nString(first_name))
        for fs in self.__first_name__:
            self.__languages__.add(fs.language)
        self.__salutation__: List[I18nString] = list()
        if isinstance(salutation, list):
            self.__salutation__.extend(salutation)
        elif salutation is not None:
            if language is not None:
                self.__salutation__.append(I18nString(salutation, lang=language))
            else:
                self.__salutation__.append(I18nString(salutation))
        for s in self.__salutation__:
            self.__languages__.add(s.language)
        return

    @property
    def last_name(self) -> I18nString:
        """
        I18nString: The last name of the individual.
        """
        return self.__last_name__

    @property
    def first_name(self) -> List[I18nString]:
        """
        List[I18nString]: The given name of the individual.
        """
        return self.__first_name__

    @property
    def salutation(self) -> List[I18nString]:
        """
        List[I18nString]: The salutation used to address an individual.
        """
        return self.__salutation__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> IndividualName | None:
        """
        Parses a lxml element into an IndividualName object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        IndividualName
            Object parsed.
        """
        if element is None:
            return None
        given_names = list()
        for given_name in element.findall('givenName', nmap):
            given_names.append(I18nString.parse(given_name, nmap))
        salutations = list()
        for salutation in element.findall('salutation', nmap):
            salutations.append(I18nString.parse(salutation, nmap))
        last_name = I18nString.parse(element.find('surName', nmap), nmap)
        individual_name = IndividualName(last_name, first_name=given_names, salutation=salutations)
        individual_name.__namespace__ = nmap
        return individual_name

    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the object.

        Returns
        -------
        lxml.tree.Element
            Object in the Element format.
        """
        individual_name = super().to_element()
        for salutation in self.salutation:
            salutation.set_tag("salutation")
            individual_name.append(salutation.to_element())
        for first_name in self.first_name:
            first_name.set_tag("givenName")
            individual_name.append(first_name.to_element())
        last_name = self.last_name
        last_name.set_tag("surName")
        individual_name.append(last_name.to_element())
        return individual_name

    def lang(self, language: Language) -> str:
        """
        Returns the individual name text representation in the given language.

        Parameters
        ----------
        language : Language
            The language to get the text representation.

        Returns
        -------
        str
            Text representation of object in the selected language
        """
        if language not in self.__languages__:
            raise AssertionError(f"{language.name.lower()} not present in {repr(self)}")
        first_name = ""
        for fs in self.first_name:
            if language == fs.language:
                first_name = f", {str(fs)[0]}."
                break
        salutation = ""
        for s in self.salutation:
            if language == s.language:
                salutation = f" {s}."
                break
        return f"{self.last_name}{first_name}{salutation}"

    def __str__(self) -> str:
        if len(self.__languages__) > 1:
            raise ValueError("Cannot convert to string when more than one language is present")
        first_name = ""
        salutation = ""
        if len(self.first_name) > 0:
            first_name = ", " + str(self.first_name[0])[0] + "."
        if len(self.salutation) > 0:
            salutation = " " + str(self.salutation[0]) + "."
        return f"{self.last_name}{first_name}{salutation}"

    def __repr__(self) -> str:
        last_name = f"Last Name: {self.last_name} [{self.last_name.language.name.lower()}]"
        first_name = ""
        if len(self.first_name) > 0:
            first_name = f""", First Name: {"/".join(
                f"{fn} [{fn.language.name.lower()}]" for fn in self.first_name
            )}"""
        salutation = ""
        if len(self.salutation) > 0:
            salutation = f""", Salutation: {"/".join(
                f"{s} [{s.language.name.lower()}]" for s in self.salutation
            )}"""
        return f"<Individual Name {{{last_name}{first_name}{salutation}}}>"



import os.path
import unittest
from unittest.mock import patch

from dwca import DarwinCoreArchive
from dwca.classes import OutsideClass
from test_dwca_classes.test_outside_class_common import TestOutsideCommon

PATH = os.path.abspath(os.path.dirname(__file__))

orig_import = __import__


def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'polars':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestOutsideNoPolars(TestOutsideCommon):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse(self, mock_import):
        super().__test_parse__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_none(self, mock_import):
        super().__test_none__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_merge(self, mock_import):
        super().__test_merge__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_missing_field(self, mock_import):
        self.assertRaises(
            AssertionError,
            OutsideClass.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
</core>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_field_twice(self, mock_import):
        self.assertRaises(
            AssertionError,
            OutsideClass.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
    <field index="2" term=""/>
    <field index="3" term=""/>
    <field index="4" term=""/>
</core>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse_invalid(self, mock_import):
        self.assertRaises(
            AssertionError,
            OutsideClass.from_string,
            '<extension />'
        )
        self.assertRaises(
            KeyError,
            OutsideClass.from_string,
            """
<core-id>
    <field index="0" term=""/>
</core-id>
            """
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def parse_read_lazy(self, mock_import):
        self.assertRaises(
            ImportError,
            DarwinCoreArchive.from_file,
            os.path.join(PATH, os.pardir, "example_data", "example_archive.zip"),
            lazy=True
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_primary_key(self, mock_import):
        self.__test_set_primary_key__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_sql_table(self, mock_import):
        self.__test_sql_table__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_insert(self, mock_import):
        self.__test_insert__()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import io
import zipfile
from copy import deepcopy
from typing import List, Dict, Type
from warnings import warn

from lxml import etree as et

from dwca.base import DarwinCore
from dwca.classes import DataFile, Occurrence, Organism, MaterialEntity, MaterialSample, Event, Location, \
    GeologicalContext, Identification, Taxon, ResourceRelationship, MeasurementOrFact, ChronometricAge, OutsideClass
from eml import EML
from eml.resources import EMLResource
from xml_common import XMLObject
from xml_common.utils import Language, read_string


class DarwinCoreArchive(DarwinCore):
    """
    Represent a Darwin Core Archive file with all its elements.

    Parameters
    ----------
    _id : str, optional
        A unique id for this Darwin Core Archive.
    """
    class Metadata(XMLObject):
        """
        Metadata class of the Darwin Core Archive storing the file name of the archive elements.

        Parameters
        ----------
        metadata : str, optional
            Name of the metadata file (e.g.: eml.xml)
        """
        PRINCIPAL_TAG = "archive"
        xmlns = "http://rs.tdwg.org/dwc/text/"
        """str : Require tag of the metadata"""
        __classes__ = [
            Occurrence, Organism, MaterialEntity, MaterialSample,
            Event, Location, GeologicalContext, Identification,
            Taxon, ResourceRelationship, MeasurementOrFact,
            ChronometricAge,
        ]

        def __init__(self, metadata: str = None) -> None:
            super().__init__()
            self.__metadata__ = metadata
            self.__core__ = None
            self.__extensions__: List[DataFile] = list()
            self.__namespace__[None] = self.xmlns
            return

        @classmethod
        def get_dwc_class(cls, element: et.Element) -> Type[DataFile]:
            """
            Extract the row type from an XML element instance.

            Parameters
            ----------
            element : lxml.etree.Element
                XML element instance.

            Returns
            -------
            Type[DataFile]
                The Python ``class`` representing the class term.
            """
            for dwc_class in cls.__classes__:
                if element.get("rowType") == dwc_class.URI:
                    return dwc_class
            warn(f"{element.get('rowType')} not in expected namespace. "
                 f"Some functionalities may not be available.")
            return OutsideClass

        @classmethod
        def parse(cls, element: et.Element, nmap: Dict) -> DarwinCoreArchive.Metadata:
            """
            Parses an `lxml.etree.Element` in a Metadata instance.

            Parameters
            ----------
            element : `lxml.etree.Element`
                XML element to be parsed.
            nmap : Dict
                Namespace prefix:uri.

            Returns
            -------
            Metadata
                New Metadata instance with the data from the element.
            """
            metadata = element.get("metadata", None)
            core_element = element.find(f"{{{nmap[None]}}}core")
            assert core_element.find("id", nmap).get("index") is not None, "Core must have an id"
            core = cls.get_dwc_class(core_element).parse(core_element, nmap=nmap)
            meta = DarwinCoreArchive.Metadata(metadata)
            meta.__core__ = core
            for extension_elem in element.findall(f"{{{nmap[None]}}}extension"):
                assert extension_elem.find("coreid", nmap).get("index") is not None, "Extension must have a coreid"
                extension = cls.get_dwc_class(extension_elem).parse(extension_elem, nmap)
                meta.__extensions__.append(extension)
            meta.__namespace__ = nmap
            return meta

        def to_element(self) -> et.Element:
            """
            Generate an element from a Metadata instance.

            Returns
            -------
            lxml.etree.Element
                XML element from Metadata instance
            """
            element = super().to_element()
            if self.__metadata__ is not None:
                element.set("metadata", self.__metadata__)
            if self.__core__ is not None:
                element.append(self.__core__.to_element())
            for extension in self.__extensions__:
                element.append(extension.to_element())
            return element

    def __init__(self, _id: str = None) -> None:
        super().__init__()
        self.__id__ = _id
        self.__meta__ = DarwinCoreArchive.Metadata()
        self.__metadata__ = None
        self.__dataset_meta__ = {
            "metadata": self.__metadata__
        }
        return

    @property
    def id(self) -> str:
        """str: A unique identifier for this DarwinCoreArchive."""
        return self.__id__

    @property
    def core(self) -> DataFile:
        """DataFile: The file with the core of the archive."""
        return self.__meta__.__core__

    @core.setter
    def core(self, core: DataFile) -> None:
        self.__meta__.__core__ = core
        try:
            df = self.__meta__.__core__.pandas
            for extension in self.extensions:
                ext_df = extension.as_pandas(_no_interaction=True)
                mask = ext_df.iloc[:, extension.id].isin(df.iloc[:, core.id])
                ext_df = ext_df.loc[ext_df[mask].index, :]
                extension.pandas = ext_df
        except ImportError:
            core_ids = [getattr(entry, core.__fields__[core.id].name) for entry in self.core.__entries__]
            for extension in self.extensions:
                def criteria(entry: DataFile.Entry) -> bool:
                    entry_id = getattr(entry, extension.__fields__[extension.id].name)
                    return entry_id in core_ids
                extension.__entries__ = list(filter(
                    criteria, extension.__entries__
                ))
        return

    @property
    def extensions(self) -> List[DataFile]:
        """
        List[DataFile]: A list with the extension of the archive.
        """
        return self.__meta__.__extensions__

    @property
    def metadata(self) -> EML:
        """
        EML: Metadata instance, currently supported EML.
        """
        return self.__metadata__

    @property
    def metadata_filename(self) -> str:
        """
        str: The filename of the metadata file.
        """
        return self.__meta__.__metadata__

    @property
    def dataset_metadata(self) -> Dict[str, EML]:
        """
        Dict[str, EML]: Metadata instances for each dataset present on DWC-A.
        """
        return self.__dataset_meta__

    @property
    def language(self) -> Language:
        """Language: Language of the Darwin Core Archive register on metadata."""
        return self.metadata.language

    def generate_eml(self, filename: str = "eml.xml") -> None:
        """
        Generate an EML file on the archive

        Parameters
        ----------
        filename : str
            Filename for the EML file to be generated
        """
        self.__meta__.__metadata__ = filename
        self.__metadata__ = EML(self.id, system="http://gbif.org", resource_type=EMLResource.DATASET)
        return

    @classmethod
    def from_file(cls, path_to_archive: str, lazy: bool = False) -> DarwinCoreArchive:
        """
        Generate a Darwin Core Archive instance from an archive file (`.zip`).

        Parameters
        ----------
        path_to_archive : str
            Path of the archive file.
        lazy : bool, optional
            Read the archive lazy. Default `False`.

        Returns
        -------
        DarwinCoreArchive
            Instance of the Darwin Core Archive.
        """
        archive = zipfile.ZipFile(path_to_archive, "r")
        index_file = archive.read("meta.xml")
        metadata = DarwinCoreArchive.Metadata.from_string(read_string(index_file))
        if metadata.__metadata__ is not None:
            metadata_content = archive.read(metadata.__metadata__)
            eml = EML.from_string(read_string(metadata_content))
            darwin_core = DarwinCoreArchive(_id=eml.package_id)
            darwin_core.__metadata__ = eml
        else:
            darwin_core = DarwinCoreArchive()
        darwin_core.__meta__ = metadata
        if lazy:
            core_file = archive.open(darwin_core.core.filename)
            darwin_core.__meta__.__core__.read_file("", source_file=core_file, lazy=lazy)
            core_file.close()
        else:
            core_file = archive.read(darwin_core.core.filename)
            darwin_core.__meta__.__core__.read_file(core_file.decode(encoding=metadata.__core__.__encoding__))
        darwin_core.__meta__.__core__._register_darwin_core_(0, darwin_core)
        for i, extension in enumerate(darwin_core.extensions):
            extension.set_core_field(darwin_core.core.__fields__[darwin_core.core.id])
            extension.set_primary_key(darwin_core.core.name)
            if lazy:
                extension_file = archive.open(extension.filename)
                darwin_core.__meta__.__extensions__[i].read_file("", source_file=extension_file, lazy=lazy)
                extension_file.close()
            else:
                extension_file = archive.read(extension.filename)
                darwin_core.__meta__.__extensions__[i].read_file(extension_file.decode())
        darwin_core.__dataset_meta__ = {
            "metadata": darwin_core.__metadata__
        }
        for item in archive.namelist():
            if item.startswith("dataset/"):
                if item == "dataset/":
                    continue
                try:
                    dataset_meta = EML.from_string(read_string(archive.read(item)))
                    darwin_core.__dataset_meta__[dataset_meta.package_id] = dataset_meta
                except Exception as e:
                    warn(f"Could not read {item.replace('dataset/', '')}:\n{e}", category=RuntimeWarning)
        archive.close()
        return darwin_core

    def to_file(
            self, path_to_archive: str,
            encoding: str = "utf-8",
            compression: int = zipfile.ZIP_DEFLATED,
            compression_level: int = 6
    ) -> None:
        """
        Generate a Darwin Core Archive file (`.zip` file) using the information of this instance.

        Parameters
        ----------
        path_to_archive : str
            Path of the archive to generate.
        encoding : str, optional
            Encoding of the corresponding files. Default `"utf-8"`.
        compression : int, optional
            The ZIP compression method to use. Default `zipfile.ZIP_DEFLATED`.
        compression_level : int, optional
            Compression level to use when writing files to the archive. Default `6`.
        """
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, 'a', compression=compression, compresslevel=compression_level) as zip_file:
            zip_file.writestr("meta.xml", self.__meta__.to_xml().encode(encoding))
            if self.metadata is not None:
                zip_file.writestr(self.__meta__.__metadata__, self.__metadata__.to_xml().encode(encoding))
            if self.core is not None:
                zip_file.writestr(self.core.filename, self.core.write_file())
            for extension in self.extensions:
                zip_file.writestr(extension.filename, extension.write_file())
            for dataset, metadata in self.dataset_metadata.items():
                if dataset != "metadata":
                    zip_file.writestr(f"dataset/{dataset}.xml", metadata.to_xml().encode(encoding))
        with open(path_to_archive, 'wb') as output_file:
            output_file.write(zip_buffer.getvalue())
        zip_buffer.close()
        return

    @classmethod
    def merge(
            cls, first_archive: DarwinCoreArchive,
            second_archive: DarwinCoreArchive,
            _id: str = None, eml: EML = None,
            eml_filename: str = "eml.xml"
    ) -> DarwinCoreArchive:
        merged_dwca = DarwinCoreArchive(_id=_id)
        if eml is not None:
            merged_dwca.__metadata__ = eml
        new_core = first_archive.core.merge(second_archive.core)
        merged_dwca.core = new_core
        first_extensions = [ext.uri for ext in first_archive.extensions]
        second_extensions = [ext.uri for ext in second_archive.extensions]
        for extension in first_archive.extensions:
            try:
                index = second_extensions.index(extension.uri)
                new_extension = extension.merge(second_archive.extensions[index])
                merged_dwca.extensions.append(deepcopy(new_extension))
                del new_extension
            except ValueError:
                merged_dwca.extensions.append(deepcopy(extension))
        for extension in second_archive.extensions:
            if extension.uri not in first_extensions:
                merged_dwca.extensions.append(deepcopy(extension))
        merged_dwca.__metadata__ = eml
        if eml is not None:
            merged_dwca.__meta__.__metadata__ = eml_filename
            merged_dwca.__dataset_meta__["metadata"] = eml
        if first_archive.metadata is not None:
            merged_dwca.__dataset_meta__[first_archive.metadata.package_id] = first_archive.metadata
        if second_archive.metadata is not None:
            merged_dwca.__dataset_meta__[second_archive.metadata.package_id] = second_archive.metadata
        for name, metadata in first_archive.dataset_metadata.items():
            if name != "metadata":
                merged_dwca.__dataset_meta__[name] = metadata
        for name, metadata in second_archive.dataset_metadata.items():
            if name != "metadata":
                merged_dwca.__dataset_meta__[name] = metadata
        return merged_dwca

    def __repr__(self) -> str:
        return f"<Darwin Core Archive ({self})>"

    def __str__(self) -> str:
        return f"{self.id} [Core: {self.core.uri}, Entries: {len(self.core)}]"



import datetime as dt
import unittest
from typing import List, Tuple, Union

from datetime_interval import Interval

from xml_common.utils import unformat_type


class NewType:
    pass


class FormatTest(unittest.TestCase):

    def test_unformat_int(self):
        a = 23
        self.assertEqual("23", unformat_type(a, int), "Incorrect encode from integer.")

    def test_unformat_float(self):
        a = 3.14
        self.assertEqual(str(3.14), unformat_type(a, float), "Incorrect encode from float.")

    def test_unformat_str(self):
        a = "23"
        self.assertEqual("23", unformat_type(a, str), "Incorrect encode from string.")
        b = "abc"
        self.assertEqual("abc", unformat_type(b, str), "Incorrect encode from string.")

    def test_unformat_list(self):
        a = ["1", "2", "3"]
        self.assertEqual("1 | 2 | 3", unformat_type(a, List[str]), "Incorrect encode from list of string.")

    def test_unformat_tuple(self):
        a = ("1", "2", "3")
        self.assertEqual(
            "1/2/3",
            unformat_type(a, Tuple[str, str, str]),
            "Incorrect encode from tuple of string."
        )
        b = (1, 2, 3)
        self.assertEqual(
            "1/2/3",
            unformat_type(b, Tuple[int, int, int]),
            "Incorrect encode from tuple of integer."
        )
        c = (1, "2", 3)
        self.assertEqual(
            "1/2/3",
            unformat_type(c, Tuple[int, str, int]),
            "Incorrect encode from tuple."
        )

    def test_unformat_datetime(self):
        a = dt.datetime(2024, 4, 5, 1, 1, 7, tzinfo=dt.timezone(dt.timedelta(hours=-6)))
        self.assertEqual(
            "2024-04-05T01:01:07-0600",
            unformat_type(a, dt.datetime),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M:%S%z`"
        )
        b = dt.datetime(2024, 4, 4, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5)))
        self.assertEqual(
            "2024-04-04T01:07-0500",
            unformat_type(b, dt.datetime),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M%z`"
        )
        c = dt.datetime(2024, 1, 1, 0, 0, 0)
        self.assertEqual(
            "2024",
            unformat_type(c, dt.datetime),
            "Incorrect format of datetime `%Y`"
        )
        d = None
        self.assertEqual(
            "", unformat_type(d, dt.datetime),
            "Incorrect format of empty datetime"
        )

    def test_unformat_interval(self):
        a = Interval(
            start=dt.datetime(2024, 4, 4, 1, 1, 7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            end=dt.datetime(2024, 4, 5, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5)))
        )
        self.assertEqual(
            "2024-04-04T01:01:07-0600/2024-04-05T01:07-0500",
            unformat_type(a, Interval),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M:%S%z/%Y-%m-%dT%H:%M%z`"
        )
        b = Interval(
            start=dt.datetime(2023, 1, 1),
            end=dt.datetime(2024, 4, 1)
        )
        self.assertEqual(
            "2023/2024-04",
            unformat_type(b, Interval),
            "Incorrect format of datetime `%Y/%Y-%m`"
        )

    def test_unformat_date_or_interval(self):
        a = Interval(
            start=dt.datetime(2024, 4, 4, 1, 1, 7, tzinfo=dt.timezone(dt.timedelta(hours=-6))),
            end=dt.datetime(2024, 4, 5, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5)))
        )
        self.assertEqual(
            "2024-04-04T01:01:07-0600/2024-04-05T01:07-0500",
            unformat_type(a, Union[Interval, dt.datetime]),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M:%S%z/%Y-%m-%dT%H:%M%z`"
        )
        b = dt.datetime(2024, 4, 4, 1, 7, 0, tzinfo=dt.timezone(dt.timedelta(hours=-5)))
        self.assertEqual(
            "2024-04-04T01:07-0500",
            unformat_type(b, Union[Interval, dt.datetime]),
            "Incorrect format of datetime `%Y-%m-%dT%H:%M%z`"
        )

    def test_incorrect_type(self):
        a = "235"
        self.assertRaisesRegex(
            AssertionError, "Value must be",
            unformat_type, a, NewType
        )
        self.assertRaisesRegex(
            AssertionError, "Value must be a list",
            unformat_type, a, List[str]
        )


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from typing import List, Union, Tuple

from dwca.terms import Field
from xml_common.utils import EstablishmentMeans


class OccurrenceID(Field):
    """
    An identifier for the Occurrence (as opposed to a particular digital record of the Occurrence).

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/occurrenceID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class CatalogNumber(Field):
    """
    An identifier (preferably unique) for the record within the data set or collection.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/catalogNumber"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RecordNumber(Field):
    """
    An identifier given to the Occurrence at the time it was recorded.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/recordNumber"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RecordedBy(Field):
    """
    A list of names of people, groups, or organizations responsible for recording the original Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/recordedBy"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class RecordedByID(Field):
    """
    A list of the globally unique identifier for the responsible for recording the original Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/recordedByID"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class IndividualCount(Field):
    """
    The number of individuals present at the time of the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/individualCount"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OrganismQuantity(Field):
    """
    A number or enumeration value for the quantity of Organisms.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismQuantity"
    TYPE = Union[int, float, str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OrganismQuantityType(Field):
    """
    The type of quantification system used for the quantity of Organisms.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismQuantityType"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OccurrenceSex(Field):
    """
    The sex of the biological individual(s) represented in the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/sex"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class LifeStage(Field):
    """
    The age class or life stage of the Organism(s) at the time the Occurrence was recorded.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/lifeStage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class ReproductiveCondition(Field):
    """
    The reproductive condition of the biological individual(s) represented in the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/reproductiveCondition"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Caste(Field):
    """
    Categorisation of individuals for eusocial species (including some mammals and arthropods).

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/caste"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Behavior(Field):
    """
    The behavior shown by the subject at the time the Occurrence was recorded.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/behavior"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Vitality(Field):
    """
    An indication of whether an Organism was alive or dead at the time of collection or observation.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/vitality"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCEstablishmentMeans(Field):
    """
    Statement about whether an Organism has been introduced through the activity of modern humans.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/establishmentMeans"
    TYPE = EstablishmentMeans

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCDegreeOfEstablishment(Field):
    """
    The degree to which an Organism survives, reproduces, and expands its range at the given place and time.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/degreeOfEstablishment"
    TYPE = str  # TODO: Create DOE Enum following https://dwc.tdwg.org/doe/

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class Pathway(Field):
    """
    The process by which an Organism came to be in a given place at a given time.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/pathway"
    TYPE = str  # TODO: Create DOE Enum following  http://rs.tdwg.org/dwc/doc/pw/.

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class GeoreferenceVerificationStatus(Field):
    """
    A categorical description of which the georeference has been verified for the Location of the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/georeferenceVerificationStatus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class OccurrenceStatus(Field):
    """
    A statement about the presence or absence of a Taxon at a Location.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    class DefaultStatus(Enum):
        PRESENT = 1
        ABSENCE = 0

    URI = "http://rs.tdwg.org/dwc/terms/occurrenceID"
    TYPE = DefaultStatus

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class AssociatedMedia(Field):
    """
    A list of identifiers (publication, global unique identifier, URI) of media associated with the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedMedia"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AssociatedOccurrences(Field):
    """
    A list of identifiers of other Occurrence records and their associations to this Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedOccurrences"
    TYPE = List[Tuple[str, str]]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AssociatedReferences(Field):
    """
    A list of identifiers of literature associated with the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedReferences"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AssociatedTaxa(Field):
    """
    A list of identifiers or names of Taxon records and the associations of this Occurrence to each of them.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedTaxa"
    TYPE = List[Tuple[str, str]]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OtherCatalogNumbers(Field):
    """
    A list of previous or alternate catalog numbers for the same Occurrence.v

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/otherCatalogNumbers"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OccurrenceRemarks(Field):
    """
    Comments or notes about the Occurrence.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/occurrenceRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



import math
import unittest

from lxml import etree as et

from xml_common.utils import LengthUnit, GPolygon, GRing
from eml.resources.coverage import GeographicCoverage
from test_xml.test_xml import TestXML


class TestGeoCoverage(TestXML):
    def test_parse(self):
        text_xml = """
<geographicCoverage id="" scope="document" system="">
    <geographicDescription>Manistee River watershed</geographicDescription>
    <boundingCoordinates>
        <westBoundingCoordinate>-118.25</westBoundingCoordinate>
        <eastBoundingCoordinate>-118.25</eastBoundingCoordinate>
        <northBoundingCoordinate>-18.25</northBoundingCoordinate>
        <southBoundingCoordinate>-118.25</southBoundingCoordinate>
        <boundingAltitudes>
            <altitudeMinimum>-10.0</altitudeMinimum>
            <altitudeMaximum>100.6</altitudeMaximum>
            <altitudeUnits>Yard_Indian</altitudeUnits>
        </boundingAltitudes>
    </boundingCoordinates>
    <datasetGPolygon>
        <datasetGPolygonOuterGRing>
            <gRing>12.0, 2.0987 12.0, -7.5555 34.345, 10.4</gRing>
        </datasetGPolygonOuterGRing>
    </datasetGPolygon>
</geographicCoverage>
        """
        geo_coverage = GeographicCoverage.from_string(text_xml)
        self.assertEqual(
            "Manistee River watershed",
            geo_coverage.description,
            "Geographic description error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.west_bounding,
            "West bounding error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.east_bounding,
            "East bounding error on parsing"
        )
        self.assertEqual(
            -18.25,
            geo_coverage.north_bounding,
            "North bounding error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.south_bounding,
            "South bounding error on parsing"
        )
        self.assertEqual(
            100.6,
            geo_coverage.max_altitude_bounding,
            "Altitude maximum bounding error on parsing"
        )
        self.assertEqual(
            -10,
            geo_coverage.min_altitude_bounding,
            "Altitude minimum bounding error on parsing"
        )
        self.assertEqual(
            LengthUnit.YARD_INDIAN,
            geo_coverage.altitude_units_bounding,
            "Altitude units bounding error on parsing"
        )
        self.assertEqual(
            [(12, 2.0987), (12, -7.5555), (34.345, 10.40)],
            geo_coverage.g_polygon[0].outer,
            "GPolygon Outer ring bounding error on parsing"
        )
        self.assertEqualTree(
            et.fromstring(text_xml),
            geo_coverage.to_element(),
            "Error on to element"
        )

    def test_parse_reference(self):
        text_xml = """
<geographicCoverage scope="system">
    <references system="http://gbif.org">1</references>
</geographicCoverage>
        """
        geo_coverage = GeographicCoverage.from_string(text_xml)
        self.assertEqual(
            "",
            geo_coverage.description,
            "Geographic description from nowhere"
        )
        self.assertTrue(
            math.isnan(geo_coverage.west_bounding),
            "West bounding from nowhere"
        )
        self.assertTrue(
            math.isnan(geo_coverage.east_bounding),
            "East bounding from nowhere"
        )
        self.assertTrue(
            math.isnan(geo_coverage.north_bounding),
            "North bounding from nowhere"
        )
        self.assertTrue(
            math.isnan(geo_coverage.south_bounding),
            "South bounding from nowhere"
        )
        self.assertIsNone(
            geo_coverage.max_altitude_bounding,
            "Altitude maximum bounding from nowhere"
        )
        self.assertIsNone(
            geo_coverage.min_altitude_bounding,
            "Altitude minimum bounding from nowhere"
        )
        self.assertIsNone(
            geo_coverage.altitude_units_bounding,
            "Altitude units bounding from nowhere"
        )
        self.assertEqual(
            0,
            len(geo_coverage.g_polygon),
            "GPolygon Outer ring bounding from nowhere"
        )
        self.assertTrue(
            geo_coverage.referencing,
            "Not referencing"
        )
        self.assertEqual(
            "1",
            geo_coverage.references,
            "Error on references"
        )
        self.assertEqual(
            "http://gbif.org",
            geo_coverage.references.system,
            "Error on references system"
        )
        self.assertEqual(
            "1",
            geo_coverage.id,
            "Error on id"
        )
        self.assertEqualTree(
            et.fromstring(text_xml),
            geo_coverage.to_element(),
            "Error on to element"
        )

    def test_parse_alternative(self):
        text_xml = """
<geographicCoverage id="" scope="document" system="">
    <geographicDescription>Manistee River watershed</geographicDescription>
    <boundingCoordinates>
        <westBoundingCoordinate>-118.25</westBoundingCoordinate>
        <eastBoundingCoordinate>-118.25</eastBoundingCoordinate>
        <northBoundingCoordinate>-18.25</northBoundingCoordinate>
        <southBoundingCoordinate>-118.25</southBoundingCoordinate>
    </boundingCoordinates>
    <datasetGPolygon>
        <datasetGPolygonOuterGRing>
            <gRingPoint>
                <gRingLatitude>34.123</gRingLatitude>
                <gRingLongitude>-118.25</gRingLongitude>
            </gRingPoint>
            <gRingPoint>
                <gRingLatitude>-18.25</gRingLatitude>
                <gRingLongitude>+25</gRingLongitude>
            </gRingPoint>
            <gRingPoint>
                <gRingLatitude>+78.25</gRingLatitude>
                <gRingLongitude>45.24755</gRingLongitude>
            </gRingPoint>
        </datasetGPolygonOuterGRing>
    </datasetGPolygon>
</geographicCoverage>
        """
        geo_coverage = GeographicCoverage.from_string(text_xml)
        self.assertEqual(
            "Manistee River watershed",
            geo_coverage.description,
            "Geographic description error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.west_bounding,
            "West bounding error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.east_bounding,
            "East bounding error on parsing"
        )
        self.assertEqual(
            -18.25,
            geo_coverage.north_bounding,
            "North bounding error on parsing"
        )
        self.assertEqual(
            -118.25,
            geo_coverage.south_bounding,
            "South bounding error on parsing"
        )
        self.assertIsNone(
            geo_coverage.max_altitude_bounding,
            "Altitude maximum bounding error on parsing"
        )
        self.assertIsNone(
            geo_coverage.min_altitude_bounding,
            "Altitude minimum bounding error on parsing"
        )
        self.assertIsNone(
            geo_coverage.altitude_units_bounding,
            "Altitude units bounding error on parsing"
        )
        self.assertEqual(
            [(34.123, -118.25), (-18.25, 25), (78.25, 45.24755)],
            geo_coverage.g_polygon[0].outer,
            "GPolygon Outer ring bounding error on parsing"
        )
        text_xml = text_xml.replace("""<gRingPoint>
                <gRingLatitude>34.123</gRingLatitude>
                <gRingLongitude>-118.25</gRingLongitude>
            </gRingPoint>
            <gRingPoint>
                <gRingLatitude>-18.25</gRingLatitude>
                <gRingLongitude>+25</gRingLongitude>
            </gRingPoint>
            <gRingPoint>
                <gRingLatitude>+78.25</gRingLatitude>
                <gRingLongitude>45.24755</gRingLongitude>
            </gRingPoint>""", """
<gRing>34.123, -118.25 -18.25, 25.0 78.25, 45.24755</gRing>
            """)
        self.assertEqualTree(
            et.fromstring(text_xml),
            geo_coverage.to_element(),
            "Error on to element"
        )

    def test_xml_to_polygon_with_points(self):
        xml = """
<datasetGPolygon>
    <datasetGPolygonOuterGRing>
        <gRingPoint>
            <gRingLatitude>34.123</gRingLatitude>
            <gRingLongitude>-118.25</gRingLongitude>
        </gRingPoint>
        <gRingPoint>
            <gRingLatitude>-18.25</gRingLatitude>
            <gRingLongitude>+25</gRingLongitude>
        </gRingPoint>
        <gRingPoint>
            <gRingLatitude>+78.25</gRingLatitude>
            <gRingLongitude>+78.25</gRingLongitude>
        </gRingPoint>
    </datasetGPolygonOuterGRing>
</datasetGPolygon>
        """
        element = et.fromstring(xml)
        polygon = GeographicCoverage.xml_to_polygon(element, {})
        self.assertEqual(
            [(34.123, -118.25), (-18.25, 25), (78.25, 78.25)],
            polygon.outer,
            "Error on parsing to polygon"
        )

    def test_xml_to_polygon_with_ring(self):
        xml = """
<datasetGPolygon>
    <datasetGPolygonOuterGRing>
        <gRing>12, 2.0987 12, -7.5555 34.345,10.40</gRing>
    </datasetGPolygonOuterGRing>
    <datasetGPolygonExclusionGRing>
        <gRing>13, 2.0987 12, -7.5555 34.345, 100</gRing>
    </datasetGPolygonExclusionGRing>
    <datasetGPolygonExclusionGRing>
        <gRing>15, 2.0987 12, -7.5555 34.345, 40</gRing>
    </datasetGPolygonExclusionGRing>
</datasetGPolygon>
        """
        element = et.fromstring(xml)
        polygon = GeographicCoverage.xml_to_polygon(element, {})
        self.assertEqual(
            [(12, 2.0987), (12, -7.5555), (34.345, 10.40)],
            polygon.outer,
            "Error on parsing to polygon (outer)"
        )
        self.assertEqual(
            GRing((13, 2.0987), (12, -7.5555), (34.345, 100)),
            polygon.exclusion[0],
            "Error on parsing to polygon (first exclusion)"
        )
        self.assertEqual(
            [(15, 2.0987), (12, -7.5555), (34.345, 40)],
            polygon.exclusion[1],
            "Error on parsing to polygon (second exclusion)"
        )
        self.assertNotEqual(
            ((15, 2.0987), (12, -7.5555), (34.345, 40)),
            polygon.exclusion[1],
            "Error on equals implementation"
        )

    def test_polygon_to_xml(self):
        xml = """
<datasetGPolygon>
    <datasetGPolygonOuterGRing>
        <gRing>12, 2.0987 12, -7.5555 34.345, 10.4</gRing>
    </datasetGPolygonOuterGRing>
    <datasetGPolygonExclusionGRing>
        <gRing>12, 2.0987 12, -7.5555 34.345, 10.4</gRing>
    </datasetGPolygonExclusionGRing>
    <datasetGPolygonExclusionGRing>
        <gRing>12, 2.0987 12, -7.5555 34.345, 10.4</gRing>
    </datasetGPolygonExclusionGRing>
</datasetGPolygon>
        """
        polygon = GPolygon(
            outer=GRing((12, 2.0987), (12, -7.5555), (34.345, 10.40)),
            exclusion=[
                GRing((12, 2.0987), (12, -7.5555), (34.345, 10.40)),
                GRing((12, 2.0987), (12, -7.5555), (34.345, 10.40)),
            ],
        )
        element = GeographicCoverage.polygon_to_xml(polygon)
        self.assertEqualTree(
            et.fromstring(xml),
            element,
            "Error on generate element from polygon"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict, List, Union, Any

from lxml import etree as et

from xml_common.utils import Language
from eml.types import EMLSection, _NoTagObject
from eml.types import I18nString


class EMLTextType(_NoTagObject):
    """
    A simple text description.

    Parameters
    ----------
    sections : List[EMLSection], optional
        A section of related text.
    paragraphs : List[I18String or str], optional
        A simple paragraph of text.
    markdowns : List[str], optional
        A block of text formatted with Markdown directives.
    language : Language, optional
        Language of abstract.
    """
    def __init__(
            self, sections: List[EMLSection] = None,
            paragraphs: List[Union[I18nString, str]] = None,
            markdowns: List[str] = None,
            language: Language = Language.ENG
    ) -> None:
        super().__init__()
        self.__sections__: List[EMLSection] = list()
        if sections is not None:
            self.__sections__.extend(sections)
        self.__paragraphs__: List[I18nString] = list()
        if paragraphs is not None:
            for paragraph in paragraphs:
                self.__paragraphs__.append(I18nString(paragraph))
        self.__markdowns__: List[str] = list()
        if markdowns is not None:
            self.__markdowns__.extend(markdowns)
        self.__lang__ = Language.ENG
        if language is not None:
            self.__lang__ = language
        return

    @property
    def sections(self) -> List[EMLSection]:
        """List[EMLSection]: A section of related text."""
        return self.__sections__

    @property
    def paragraphs(self) -> List[I18nString]:
        """List[I18nString]: A simple paragraph of text."""
        return self.__paragraphs__

    @property
    def markdowns(self) -> List[str]:
        """List[str]: A block of text formatted with Markdown directives."""
        return self.__markdowns__

    @property
    def language(self) -> Language:
        """Language: Language of the text."""
        return self.__lang__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLTextType | None:
        """
        Generate an EMLTextType from an XML element.
        
        Parameters
        ----------
        element : lxml.etree.Element
            XML Element to be parsed.
        nmap : Dict
            Namespace of the XML element.

        Returns
        -------
        EMLTextType
            Parsed object
        """
        if element is None:
            return None
        sections = list()
        for section in element.findall("section", nmap):
            sections.append(EMLSection.parse(section, nmap))
        paragraphs = list()
        for paragraph in element.findall("para", nmap):
            paragraphs.append(I18nString.parse(paragraph, nmap))
        markdowns = list()
        for markdown in element.findall("markdown", nmap):
            markdowns.append(markdown.text)
        text_type = EMLTextType(sections, paragraphs, markdowns)
        text_type.__namespace__ = nmap
        return text_type

    def to_element(self) -> et.Element:
        """
        Get an XML Element from the EMLTextType object

        Returns
        -------
        lxml.etree.Element
            XML Element with the information of the EMLTextType instance
        """
        text_type = super().to_element()
        for section in self.sections:
            text_type.append(section.to_element())
        for para in self.paragraphs:
            para.set_tag("para")
            text_type.append(para.to_element())
        for markdown in self.markdowns:
            mark_elem = self.object_to_element("markdown")
            mark_elem.text = markdown
            text_type.append(mark_elem)
        return text_type

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EMLTextType):
            return (
                self.sections == other.sections and
                self.paragraphs == other.paragraphs and
                self.markdowns == other.markdowns
            )
        else:
            return False



import datetime as dt
from datetime import datetime
from datetime_interval import Interval
from typing import TypeAlias, get_args, List, Any, Union, get_origin
from warnings import warn

POSSIBLE_DATETIME_FORMATS = [
    "%Y-%m-%dT%H:%M:%S%z",
    "%Y-%m-%dT%H:%M%z",
    "%Y-%m-%dT%H%z",
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%dT%H:%M",
    "%Y-%m-%dT%H",
    "%Y-%m-%d",
    "%Y-%m",
    "%Y",
]


def format_to_type(value: str, a_type: TypeAlias, address_value: bool = True) -> Any:
    """
    Convert the string value according to the type given.

    Parameters
    ----------
    value : str
        A value to be converted.
    a_type : TypeAlias
        The type to convert the value.
    address_value : bool, optional
        If the conversion gives ValueError, whether return None.

    Returns
    -------
    Any
        The value as a TypeAlias.

    Raises
    ------
    TypeError
        When the Type alias is not supported or value cannot be converted in that particular Type.
    """
    try:
        if a_type == Interval:
            raise TypeError
        return a_type(value)
    except ValueError as e:
        if address_value:
            return None
        else:
            raise e
    except TypeError as e:
        if a_type == List[str]:
            if value is None:
                return None
            return value.split(" | ")
        elif a_type.__name__ == "Tuple":
            return tuple([
                format_to_type(
                    this_value,
                    this_type
                ) for this_value, this_type in zip(
                    value.split("/"),
                    get_args(a_type)
                )
            ])
        elif a_type == Any:
            warn("<Any> type is not recommended as Type of field.")
            return value
        elif a_type == Interval:
            start_value, end_value = value.split("/")
            return Interval(start=format_datetime(start_value), end=format_datetime(end_value))
        elif a_type == dt.datetime:
            return format_datetime(value)
        elif getattr(a_type, "__origin__", None) is Union:
            return format_union(value, a_type)
        else:
            if str(a_type)[0] == "<":
                type_str = str(a_type)
            else:
                type_str = f"<{a_type}>"
            raise TypeError(f"Type {type_str} does not have automatic conversion.")


def format_union(value: str, types: TypeAlias) -> Any:
    """
    Convert the string value to any of the types given inside the union.

    Parameters
    ----------
    value : str
        A value to be converted.
    types : TypeAlias
        A Union Type with various option of types.

    Returns
    -------
    Any
        Value in one of the type listed inside the `types` parameter.

    Raises
    ------
    TypeError
        When the value cannot be converted in any of Types given.
    """
    previous_exception = None
    for a_type in get_args(types):
        try:
            return format_to_type(value, a_type, address_value=False)
        except (TypeError, ValueError) as e:
            e.__cause__ = previous_exception
            previous_exception = e
            continue
    type_str = [f"<{a_type}>" if str(a_type)[0] != "<" else str(a_type) for a_type in get_args(types)]
    exception = TypeError(f"{value} does not match any of {', '.join(type_str)}")
    exception.__cause__ = previous_exception
    raise exception


def format_datetime(value: str) -> dt.datetime | None:
    """
    Convert value in a datetime object.

    Parameters
    ----------
    value : str
        Value to be parsed in a datetime object.

    Returns
    -------
    datatime
        Datetime object from value.
    """
    previous_exception = None
    if value == "" or value is None:
        return
    for dt_format in POSSIBLE_DATETIME_FORMATS:
        try:
            return dt.datetime.strptime(value.strip(), dt_format)
        except ValueError as e:
            e.__cause__ = previous_exception
            previous_exception = e
            continue
    format_string = ", \n".join(POSSIBLE_DATETIME_FORMATS)
    exception = ValueError(f"{value} does not match any of:\n{format_string}")
    exception.__cause__ = previous_exception
    raise exception


def unformat_type(value: Any, a_type: TypeAlias) -> str:
    """
    Convert a value to a string according to the type given.

    Parameters
    ----------
    value : Any
        A value in a_type format.
    a_type : TypeAlias
        The type of the value given.

    Returns
    -------
    str
        Encoded value
    """
    try:
        import pandas as pd
        if pd.isna(value):
            return ""
    except ImportError:
        pass
    except ValueError:
        pass
    if value is None:
        return ""
    if a_type == List[str]:
        assert isinstance(value, list), f"Value must be a list of string"
        return " | ".join([str(v) for v in value])
    elif a_type.__name__ == "Tuple":
        encoded_str = [unformat_type(v, this_type) for v, this_type in zip(value, get_args(a_type))]
        return "/".join(encoded_str)
    else:
        assert isinstance(value, a_type), f"Value must be an instance of {a_type}"
        if a_type == Interval or (Interval in get_args(a_type) and isinstance(value, Interval)):
            return f"{unformat_datetime(value.start)}/{unformat_datetime(value.end)}"
        elif a_type == dt.datetime or (dt.datetime in get_args(a_type) and isinstance(value, dt.datetime)):
            return unformat_datetime(value)
        else:
            return str(value)


def unformat_datetime(value: dt.datetime) -> str:
    """
    Convert a datetime value in an encoded string.

    Parameters
    ----------
    value : datetime
        Value to be encoded as a string.

    Returns
    -------
    str
        Encoded Datetime object.
    """
    if value.tzinfo is None:
        z = ""
    elif value.tzinfo == dt.timezone.utc:
        z = "Z"
    else:
        z = "%z"
    if value.second == 0:
        if value.minute == 0:
            if value.hour == 0:
                if value.day == 1:
                    if value.month == 1:
                        return value.strftime(f"%Y")
                    else:
                        return value.strftime(f"%Y-%m")
                else:
                    return value.strftime(f"%Y-%m-%d")
            else:
                return value.strftime(f"%Y-%m-%dT%H{z}")
        else:
            return value.strftime(f"%Y-%m-%dT%H:%M{z}")
    else:
        return value.strftime(f"%Y-%m-%dT%H:%M:%S{z}")


def type_to_pl(a_type: TypeAlias, lazy: bool = False) -> TypeAlias:
    """
    Equivalent to a_type in the polars dtype.

    Parameters
    ----------
    a_type : TypeAlias
        Any available type.
    lazy : bool
        When used to inference the types in lazy mode.

    Returns
    -------
    TypeAlias
        polars dtype.
    """
    try:
        import polars as pl
    except ImportError:
        raise ImportError("polars not installed.")
    if a_type == str:
        return pl.String
    elif get_origin(a_type) == list:
        if lazy:
            return pl.String
        inner_type = get_args(a_type)
        if len(inner_type) == 1:
            return pl.List(type_to_pl(get_args(a_type)[0]))
        else:
            return pl.List(pl.Object)
    elif a_type == bool:
        return pl.Boolean
    elif a_type == int:
        return pl.Int64
    elif a_type == float:
        return pl.Float64
    elif a_type == dt.datetime:
        return pl.Datetime
    else:
        if lazy:
            return pl.String
        return pl.Object

def type_to_sql(a_type: TypeAlias) -> str:
    """
    Equivalent to a_type in the polars dtype.

    Parameters
    ----------
    a_type : TypeAlias
        Any available type.

    Returns
    -------
    TypeAlias
        polars dtype.
    """
    if a_type == int:
        return "INTEGER"
    elif a_type == float:
        return "REAL"
    elif a_type == str:
        return "VARCHAR"
    elif a_type == bool:
        return "BOOLEAN"
    elif a_type == bytes:
        return "BLOB"
    elif a_type == dt.datetime:
        return "DATETIME"
    else:
        return "VARCHAR"


def format_to_sql(value: Any, a_type: TypeAlias) -> Any:
    """
    Format a value to the SQL equivalent.

    Parameters
    ----------
    value : Any
        Value to be formatted.
    a_type : TypeAlias
        The type of the value.

    Returns
    -------
    Any
        Value formatted.
    """
    given_type = type_to_sql(a_type)
    if value is None:
        return None
    if given_type == "INTEGER":
        return int(value)
    elif given_type =="REAL":
        return float(value)
    elif given_type == "BOOLEAN":
        return bool(value)
    elif given_type == "BLOB":
        if isinstance(value, bytes):
            return value
        elif isinstance(value, str):
            return bytes(value, encoding="ascii")
        else:
            return bytes(value)
    elif given_type == "DATETIME":
        if isinstance(value, str):
            return format_datetime(value)
        elif isinstance(value, dt.datetime):
            return value
        else:
            raise ValueError(f"{type(value)} is not supported to datetime.")
    elif given_type == "VARCHAR":
        return str(value)
    else:
        return str(value)



import unittest

from lxml import etree as et

from eml.types import SemanticAnnotation
from test_xml.test_xml import TestXML


class TestSemanticAnnotation(TestXML):
    DEFAULT_TAGS = {
        "scope": "document"
    }

    def setUp(self) -> None:
        self.text_xml = """
        <annotation id="Annotation">
            <propertyURI 
                label="has characteristic"
            >http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#ofCharacteristic</propertyURI>
            <valueURI
                label="Mass"
            >http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#Mass</valueURI>
        </annotation>
                """
        self.annotation = SemanticAnnotation.from_string(self.text_xml)

    def test_parse(self):
        self.assertEqual(
            "http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#ofCharacteristic",
            self.annotation.property_uri,
            "Error on parsing property URI"
        )
        self.assertEqual(
            "has characteristic",
            self.annotation.property_uri.label,
            "Error on parsing property URI label"
        )
        self.assertEqual(
            "http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#Mass",
            self.annotation.value_uri,
            "Error on parsing value URI"
        )
        self.assertEqual(
            "Mass",
            self.annotation.value_uri.label,
            "Error on parsing value URI label"
        )
        self.assertRaises(
            RuntimeError,
            self.annotation.to_element
        )
        self.annotation.set_tag("annotation")
        self.assertEqualTree(et.fromstring(self.text_xml), self.annotation.to_element(), "Error on to element")

    def test_referencing(self):
        text_xml = """
<annotation>
    <references>Invalid</references>
</annotation>
        """
        self.assertRaises(
            ValueError, SemanticAnnotation.from_string, text_xml
        )

    def test_equal(self):
        annotation = SemanticAnnotation.from_string(self.text_xml)
        self.assertEqual(
            self.annotation,
            annotation,
            "Annotation not equal"
        )
        self.assertNotEqual(
            1,
            annotation,
            "Annotation equal to an integer"
        )



if __name__ == '__main__':
    unittest.main()



import unittest
from lxml import etree as et

from eml.types import EMLPhone
from test_xml.test_xml import TestXML


class TestEMLPhone(TestXML):
    EXAMPLE_XML = """
<phone phonetype="voice">805-555-2500</phone>
    """

    def test_parse(self):
        eml_phone = EMLPhone.from_string(self.EXAMPLE_XML)
        self.assertEqual("805-555-2500", eml_phone, "Phone set incorrectly")
        self.assertEqual("voice", eml_phone.phone_type, "Phonetype set incorrectly")
        self.assertEqualTree(et.fromstring(self.EXAMPLE_XML), eml_phone.to_element(), "Incorrect generated element")
        return

    def test_equal(self):
        variable = EMLPhone("805-555-2500")
        self.assertEqual(EMLPhone("805-555-2500"), variable, "Phones not equal")
        self.assertNotEqual(8055552500, variable, "Phones equal to other type")
        self.assertNotEqual(EMLPhone("8055552500"), variable, "Phones equal to different phone")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict, Union

from lxml import etree as et

from xml_common.utils import Language
from eml.types import _NoTagObject


class I18nString(_NoTagObject):
    """
    Internationalization of a string representing a text and its language.

    Parameters
    ----------
    value : str or I18nString
        Value of the string.
    lang : Language or str
        Language of the string or valid abbreviation. Default `ENG` (English).

    Raises
    ------
    TypeError
        Wrong language type.
    NotImplementedError
        Abbreviation not available yet or wrong language.
    """
    def __init__(self, value: Union[str, I18nString], lang: Union[Language, str] = Language.ENG) -> None:
        super().__init__()
        if isinstance(value, I18nString):
            self.__value__ = str(value)
            self.__lang__ = value.language
            self.__tag__ = value.__tag__
        else:
            if value is None:
                self.__value__ = ""
            else:
                self.__value__ = value
            language = Language.ENG
            if lang is None:
                pass
            elif isinstance(lang, str):
                found = False
                for candid_lang in Language:
                    if lang.lower() == candid_lang.name.lower():
                        language = candid_lang
                        found = True
                        break
                if not found:
                    raise NotImplementedError(f"{lang} not in the available languages")
            elif isinstance(lang, Language):
                language = lang
            else:
                raise TypeError("Language must be a Language type or a valid string")
            self.__lang__ = language
        return

    @property
    def language(self) -> Language:
        """Language: The language of the string."""
        return self.__lang__

    def __str__(self) -> str:
        return self.__value__

    def __repr__(self) -> str:
        return f"<i18n String ({self.__value__}) [{self.language.name.lower()}]>"

    def __eq__(self, other: Union[str, I18nString]) -> bool:
        if isinstance(other, str):
            return self.__value__ == other
        elif isinstance(other, I18nString):
            return self.__value__ == other.__value__ and self.language == other.language
        else:
            return False

    def __gt__(self, other: Union[str, I18nString]) -> bool:
        if isinstance(other, str):
            return self.__value__ > other
        elif isinstance(other, I18nString):
            return str(self) > str(other)
        else:
            raise TypeError(f"'>' not supported between instances of 'I18n String' and '{type(other)}'")

    def __ge__(self, other: Union[str, I18nString]) -> bool:
        return self.__value__ > other or self.__value__ == other

    def __lt__(self, other: Union[str, I18nString]) -> bool:
        if isinstance(other, str):
            return self.__value__ < other
        elif isinstance(other, I18nString):
            return str(self) < str(other)
        else:
            raise TypeError(f"'<' not supported between instances of 'I18n String' and '{type(other)}'")

    def __le__(self, other: Union[str, I18nString]) -> bool:
        return self.__value__ < other or self.__value__ == other

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> I18nString | None:
        """
        Parses a lxml element into an i18n String.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        I18nString
            Object parsed.
        """
        if element is None:
            return None
        text = element.text if element.text is not None else ""
        return I18nString(text, lang=element.get(f"{{{cls.NAMESPACES['xml']}}}lang", "eng"))

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the i18n string.

        Returns
        -------
        lxml.tree.Element
            Object in the XML Element format.

        Raises
        ------
        RuntimeError
            If tag is not set before calling this method.
        """
        element = super().to_element()
        element.text = self.__value__
        element.set(f"{{{self.__namespace__['xml']}}}lang", self.__lang__.name.lower())
        return element



from __future__ import annotations

from typing import Dict

from lxml import etree as et

from eml.resources.distribution import EMLOnline, EMLOffline, EMLInline
from eml.types import EMLObject, Scope


class EMLDistribution(EMLObject):
    """
    Information on how the resource is distributed online and offline.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    online : EMLOnline, optional
        Online distribution information.
    offline : EMLOffline, optional
        Data are available offline.
    inline : EMLInline, optional
        Data distributed inline in the metadata.
    """
    PRINCIPAL_TAG = "distribution"

    def __init__(self, _id: str = None, scope: Scope = Scope.DOCUMENT, system: str = None, referencing: bool = False,
                 references_system: str = None, online: EMLOnline = None, offline: EMLOffline = None,
                 inline: EMLInline = None) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        if not self.referencing:
            declared = (online is not None) + (offline is not None) + (inline is not None)
            if declared == 0:
                raise TypeError("At least declare online, offline or inline distribution.")
            if declared > 1:
                raise TypeError("Declare only one of the following: online, offline or inline distribution.")
        self.__online__ = online
        self.__offline__ = offline
        self.__inline__ = inline
        return

    @property
    def online(self) -> EMLOnline:
        """EMLOnline: Online distribution information."""
        return self.__online__

    @property
    def offline(self) -> EMLOffline:
        """EMLOffline: Data are available offline."""
        return self.__offline__

    @property
    def inline(self) -> EMLInline:
        """EMLInline: Data distributed inline in the metadata."""
        return self.__inline__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLDistribution:
        """
        Generate an EML Distribution referencing another EML Distribution.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLDistribution
            Object parsed that reference another.
        """
        references = element.find("references", nmap)
        return EMLDistribution(
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLDistribution:
        """
        Generate an EML Distribution that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLObject
            Object parsed.
        """
        online = EMLOnline.parse(element.find("online", nmap), nmap)
        offline = EMLOffline.parse(element.find("offline", nmap), nmap)
        inline = EMLInline.parse(element.find("inline", nmap), nmap)
        return EMLDistribution(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            online=online,
            offline=offline,
            inline=inline,
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element with EMLDistribution instance information.

        Returns
        -------
        lxml.etree.Element
            XML element object
        """
        dist_elem = super().to_element()
        dist_elem = self._to_element_(dist_elem)
        dist_elem.append(self.online.to_element()) if self.online is not None else None
        dist_elem.append(self.offline.to_element()) if self.offline is not None else None
        dist_elem.append(self.inline.to_element()) if self.inline is not None else None
        return dist_elem



from __future__ import annotations

from typing import List, Tuple, Any, Iterator


class GRing:
    """
    A set of ordered pairs of floating-point numbers.

    Parameters
    ----------
    points : Tuple[float, float]
        Points to generate a ring, minimum 3 points with latitude and longitude.
    """
    def __init__(self, *points: Tuple[float, float]) -> None:
        self.__points__ = list()
        assert len(points) >= 3, "At least 3 points to generate a ring"
        for point in points:
            self.__points__.append(point)
        return

    def __iter__(self) -> Iterator[Tuple[float, float]]:
        for point in self.__points__:
            yield point

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, GRing):
            return sorted(self.__points__) == sorted(other.__points__)
        if isinstance(other, list):
            return sorted(self.__points__) == sorted(other)
        return False


class GPolygon:
    """
    This construct creates a spatial ring with a hollow center.

    Parameters
    ----------
    outer : GRing
        The outer containment loop of a datasetGPolygon.
    exclusion : List[GRing], optional
        Exclusion G-Ring, the closed non-intersecting boundary of a void area (or hole in an interior area).
    """
    def __init__(
            self,
            outer: GRing,
            exclusion: List[GRing] = None
    ) -> None:
        self.__outer__ = outer
        self.__exclusion__ = list()
        if exclusion is not None:
            self.__exclusion__.extend(exclusion)
        return

    @property
    def outer(self) -> GRing:
        """GRing: The outer containment loop of a datasetGPolygon."""
        return self.__outer__

    @property
    def exclusion(self) -> List[GRing]:
        """List[GRing]: Exclusion G-Ring."""
        return self.__exclusion__



from __future__ import annotations

from dwca.base import DarwinCore


class SimpleDarwinCore(DarwinCore):
    """
    Class representing a Simple Darwin Core standard.
    """
    @classmethod
    def from_file(cls, path_to_archive: str) -> SimpleDarwinCore:
        pass

    def to_file(self, path_to_archive: str, encoding: str = "utf-8") -> None:
        pass



import datetime as dt
from typing import Any, Dict, Union

from datetime_interval import Interval

from dwca.terms import Field
from xml_common.utils import Language


class DWCType(Field):
    """
    The nature or genre of the resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/elements/1.1/type"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCModified(Field):
    """
    The most recent date-time on which the resource was changed.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/modified"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCLanguage(Field):
    """
    A language of the resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/elements/1.1/language"
    TYPE = Language

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return

    def format(self, value: str) -> Language:
        """
        Format value to Language.

        Parameters
        ----------
        value : str
            Value to be formatted as a Language.

        Returns
        -------
        Language
            Value as a Language object.
        """
        return Language.get_language(value)

    def unformat(self, value: Language) -> str:
        """
        Encode value from Language to a standard string.

        Parameters
        ----------
        value : Language
            An instance of Language to be represented as a string.

        Returns
        -------
        str
            String representation of Language.
        """
        return value.name.lower()


class DWCLicense(Field):
    """
    A legal document giving official permission to do something with the resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/license"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCRightsHolder(Field):
    """
    A person or organization owning or managing rights over the resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/rightsHolder"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCAccessRights(Field):
    """
    Information about who can access the resource or an indication of its security status.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/accessRights"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCBibliographicCitation(Field):
    """
    A bibliographic reference for the resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/bibliographicCitation"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCReferences(Field):
    """
    A related resource that is referenced, cited, or otherwise pointed to by the described resource.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://purl.org/dc/terms/references"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCInstitution(Field):
    """
    An identifier for the institution having custody of the object(s) or information referred to in the record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/institutionID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCCollection(Field):
    """
    An identifier for the collection or dataset from which the record was derived.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/collectionID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCDataset(Field):
    """
    An identifier for the set of data.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/datasetID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCInstitutionCode(Field):
    """
    The name (or acronym) in use by the institution having custody of the information referred to in the record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/institutionCode"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCCollectionCode(Field):
    """
    The name, acronym, coden, or initialism identifying the collection or data set from which the record was derived.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/collectionCode"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCDatasetName(Field):
    """
    The name identifying the data set from which the record was derived.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/datasetName"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCOwnerInstitutionCode(Field):
    """
    The name (or acronym) in use by the institution having ownership of the object(s) or information referred to in the record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/ownerInstitutionCode"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCBasisOfRecord(Field):
    """
    The specific nature of the data record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/basisOfRecord"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCInformationWithheld(Field):
    """
    Additional information that exists, but that has not been shared in the given record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/informationWithheld"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCDataGeneralizations(Field):
    """
    Actions taken to make the shared data less specific or complete than in its original form.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/dataGeneralizations"
    TYPE = Any

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class DWCDynamicProperties(Field):
    """
    A list of additional measurements, facts, characteristics, or assertions about the record.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/dynamicProperties"
    TYPE = Dict[str, Any]

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return



from __future__ import annotations

from typing import Any, Dict
from lxml import etree as et

from dwca.terms import Field


class OutsideTerm(Field):
    """
    Terms defined outside the Darwin Core specifications.

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    uri : str
        URI of the term.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    TYPE = str

    def __init__(self, index: int | str, uri: str = None, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        self.URI = uri
        return

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> OutsideTerm | None:
        """
        Generate a OutsideTerm object from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element instance.
        nmap : Dict
            Namespace.

        Returns
        -------
        OutsideTerm
            An OutsideTerm object.
        """
        if element is None:
            return None
        if "term" not in element.attrib:
            raise TypeError("Field must have a term")
        field = OutsideTerm(
            element.get("index"),
            element.get("term", None),
            element.get("default", None),
            element.get("vocabulary", None),
        )
        field.__namespace__ = nmap
        return field



from __future__ import annotations
from typing import Dict, Union, List, Tuple

from lxml import etree as et

from eml.types import EMLObject, _NoTagObject, Scope, ResponsibleParty, Role, EMLTextType


class ResearchProject(EMLObject, _NoTagObject):
    """
    Descriptor of a research context for a dataset or another project.

    Parameters
    ----------
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    title : str | List[str]
        Title(s) of the project.
    personnel : List[Tuple[ResponsibleParty, Role]]
        Contact and role information for people involved in the research project.
    abstract : EMLTextType, optional
        Project Abstract.
    funding : EMLTextType, optional
        Funding information.
    related_projects : List[ResearchProject], optional
        This field is a recursive link to another project.

    Raises
    ------
    TypeError
        If it is not a reference, raises TypeError if title or personnel is not given.
    """
    def __init__(
            self, _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None,
            title: Union[str, List[str]] = None,
            personnel: List[Tuple[ResponsibleParty, Role]] = None,
            abstract: EMLTextType = None,
            funding: EMLTextType = None,
            related_projects: List[ResearchProject] = None
    ) -> None:
        EMLObject.__init__(self, _id, scope, system, referencing, references_system)
        _NoTagObject.__init__(self)
        self.__titles__ = list()
        self.__personnel__ = list()
        self.__abstract__ = abstract
        self.__funding__ = funding
        self.__related_projects__ = list()
        if not self.referencing:
            if title is None or len(title) == 0:
                raise TypeError("ResearchProject requires at least one title.")
            if personnel is None or len(personnel) == 0:
                raise TypeError("ResearchProject requires at least one personnel.")
            if isinstance(title, str):
                self.__titles__.append(title)
            else:
                self.__titles__.extend(title)
            self.__personnel__ = personnel
            if related_projects is not None:
                self.__related_projects__.extend(related_projects)
        return

    @property
    def title(self) -> str | None:
        """str: A descriptive title for the research project."""
        if len(self.__titles__) == 0:
            return None
        return self.__titles__[0]

    @property
    def titles(self) -> List[str]:
        """List[str]: A list of all titles of this project."""
        return self.__titles__

    @property
    def personnel(self) -> List[Tuple[ResponsibleParty, Role]]:
        """List[Tuple[ResponsibleParty, Role]]: Contact and role information for people involved in the research project."""
        return self.__personnel__

    @property
    def abstract(self) -> EMLTextType:
        """EMLTextType: Project Abstract."""
        return self.__abstract__

    @property
    def funding(self) -> EMLTextType:
        """EMLTextType: Funding information."""
        return self.__funding__

    @property
    def award(self) -> List:
        """List: Awards information."""
        # TODO: Implement Awards (https://eml.ecoinformatics.org/schema/eml-project_xsd.html#ResearchProjectType_award)
        return list()

    @property
    def study_area_description(self) -> None:
        """None: Description of the study area."""
        # TODO: Implement StudyAreaDescription (https://eml.ecoinformatics.org/schema/eml-project_xsd.html#ResearchProjectType_studyAreaDescription)
        return

    @property
    def design_description(self) -> None:
        """None: Description of research design."""
        # TODO: Implement DesignDescription (https://eml.ecoinformatics.org/schema/eml-project_xsd.html#ResearchProjectType_designDescription)
        return

    @property
    def related_project(self) -> List[ResearchProject]:
        """List[ResearchProject]: List of related projects."""
        return self.__related_projects__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> ResearchProject:
        """
        Generate a Research Project referencing another Research Project.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        ResearchProject
            Object parsed that reference another research project on document.
        """
        references = element.find("references")
        return ResearchProject(
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> ResearchProject:
        """
        Generate a Research Project that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        ResearchProject
            Object parsed.
        """
        abstract_elem = element.find("abstract", namespaces=nmap)
        related_elem = element.findall("relatedProject", namespaces=nmap)
        return ResearchProject(
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
            title=[title.text for title in element.findall("title")],
            personnel=[
                (ResponsibleParty.parse(person, nmap), ResponsibleParty.get_role(person, nmap))
                for person in element.findall("personnel")
            ],
            abstract=None if abstract_elem is None else EMLTextType.parse(abstract_elem, nmap),
            related_projects=[ResearchProject.parse(related, nmap) for related in related_elem]
        )

    def to_element(self) -> et.Element:
        """
        Generate a lxml.tree.Element from the Research Project object.

        Returns
        -------
        lxml.tree.Element
            Object in the XML Element format.

        Raises
        ------
        RuntimeError
            If tag is not set before calling this method.
        """
        res_proj = _NoTagObject.to_element(self)
        res_proj = self._to_element_(res_proj)
        references = self.generate_references_element()
        if references is None:
            for title in self.titles:
                title_elem = self.object_to_element("title")
                title_elem.text = title
                res_proj.append(title_elem)
            for person, role in self.personnel:
                person.set_tag("personnel")
                person_elem = person.to_element()
                role_elem = self.object_to_element("role")
                role_elem.text = role.to_camel_case()
                person_elem.append(role_elem)
                res_proj.append(person_elem)
            for related in self.related_project:
                related.set_tag("relatedProject")
                res_proj.append(related.to_element())
        return res_proj



from eml.resources.resources import EMLResource
from eml.resources.keyword_set import EMLKeywordSet, KeywordType
from eml.resources.licence import EMLLicense
from eml.resources.distribution import EMLDistribution
from eml.resources.coverage import EMLCoverage
from eml.resources.resource import Resource
from eml.resources.dataset import EMLDataset
from eml.resources.citation import EMLCitation
from eml.resources.protocol import EMLProtocol
from eml.resources.software import EMLSoftware
